# -*- coding: utf-8 -*-
"""
users manager module.
"""

from pyrin.security.users.manager import UsersManager as BaseUsersManager
from pyrin.core.exceptions import CoreNotImplementedError


class UsersManager(BaseUsersManager):
    """
    users manager class.
    """

    def is_active(self, user, **options):
        """
        gets a value indicating that given user is active.

        :param user: user to check its active status.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
database orm sql extractor manager module.
"""

import sqlparse

from sqlalchemy.sql.elements import TextClause
from sqlparse.tokens import DML, Keyword
from sqlparse.sql import IdentifierList, Identifier, Function, Where, Comparison, TokenList

from pyrin.core.context import Manager
from pyrin.core.globals import LIST_TYPES


class DatabaseORMSQLExtractorManager(Manager):
    """
    database orm sql extractor manager class.
    """

    def find_table_names(self, expression, **options):
        """
        finds table names from a string or `TextClause` sql expression.

        :param Union[str, TextClause] expression: a string or `TextClause`
                                                  containing a sql expression.

        :keyword bool include_select: specifies that select statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_insert: specifies that insert statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_update: specifies that update statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_delete: specifies that delete statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :returns: list[str]
        :rtype: list
        """

        sql = expression
        tables = []
        if isinstance(expression, TextClause):
            sql = expression.text

        if sql is None or len(sql) <= 0:
            return tables

        return self._extract_tables(sql, **options)

    def _is_subselect(self, statement):
        """
        gets a value indicating that given statement is a select or sub-select.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'SELECT':
                return True
        return False

    def _is_insert(self, statement):
        """
        gets a value indicating that given statement is an insert.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'INSERT':
                return True
        return False

    def _is_update(self, statement):
        """
        gets a value indicating that given statement is an update.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'UPDATE':
                return True
        return False

    def _is_delete(self, statement):
        """
        gets a value indicating that given statement is a delete.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'DELETE':
                return True
        return False

    def _is_join(self, token):
        """
        gets a value indicating that given token represents a join keyword.

        :param Token token: token to be checked for join.

        :rtype: bool
        """

        if token.ttype is Keyword and (token.value.upper() == 'JOIN' or
                                       ' JOIN' in token.value.upper() or
                                       token.value.upper() == 'ON'):
            return True

        return False

    def _is_where(self, token):
        """
        gets a value indicating that given token represents a where clause.

        :param Token token: token to be checked for where clause.

        :rtype: bool
        """

        return isinstance(token, Where)

    def _is_comparison(self, token):
        """
        gets a value indicating that given token represents a comparison clause.

        :param Token token: token to be checked for comparison clause.

        :rtype: bool
        """

        return isinstance(token, Comparison)

    def _is_table(self, token):
        """
        gets a value indicating that given token represents a table.

        :param Token token: token to be checked for being table.

        :rtype: bool
        """

        return not self._is_join(token) and not self._is_where(token) and \
            not self._is_comparison(token)

    def _extract_select_from_part(self, statement):
        """
        extracts table names from a select statement.
        it also supports multiple joins and sub-selects.

        :param Statement statement: statement object.

        :returns: list[str]
        :rtype: list
        """

        select_seen = False
        from_seen = False
        join_seen = False
        results = []
        for item in statement.tokens:
            if item.is_group:
                for x in self._extract_select_from_part(item):
                    results.append(self._get_identifier(x, True))
            if select_seen is True and from_seen is True and join_seen is False:
                if self._is_subselect(item):
                    for z in self._extract_select_from_part(item):
                        results.append(self._get_identifier(z, True))
                elif item.ttype is Keyword and item.value.upper() in ['ORDER BY',
                                                                      'GROUP BY',
                                                                      'HAVING']:
                    return results
                elif self._is_join(item) is True:
                    join_seen = True
                elif item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    results.append(self._get_identifier(item, True))
            elif select_seen is True and from_seen is True and join_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    results.append(self._get_identifier(item, True))
            elif item.ttype is Keyword.DML and item.value.upper() == 'SELECT':
                select_seen = True
            elif item.ttype is Keyword and item.value.upper() == 'FROM':
                from_seen = True
            elif self._is_join(item) is True:
                join_seen = True

        return results

    def _extract_insert_into_part(self, statement):
        """
        extracts table name from an insert statement.

        :param Statement statement: statement object.

        :rtype: str
        """

        insert_seen = False
        into_seen = False
        for item in statement.tokens:
            if insert_seen is True and into_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    return self._get_identifier(item, True)
            elif item.ttype is Keyword.DML and item.value.upper() == 'INSERT':
                insert_seen = True
            elif item.ttype is Keyword and item.value.upper() == 'INTO':
                into_seen = True

        return None

    def _extract_update_part(self, statement):
        """
        extracts table name from an update statement.

        :param Statement statement: statement object.

        :rtype: str
        """

        update_seen = False
        for item in statement.tokens:
            if update_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    return self._get_identifier(item, True)
            elif item.ttype is Keyword.DML and item.value.upper() == 'UPDATE':
                update_seen = True

        return None

    def _extract_delete_from_part(self, statement):
        """
        extracts table name from a delete statement.

        :param Statement statement: statement object.

        :rtype: str
        """

        delete_seen = False
        from_seen = False
        for item in statement.tokens:
            if delete_seen is True and from_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    return self._get_identifier(item, True)
            elif item.ttype is Keyword.DML and item.value.upper() == 'DELETE':
                delete_seen = True
            elif item.ttype is Keyword and item.value.upper() == 'FROM':
                from_seen = True

        return None

    def _extract_table_identifiers(self, token_stream):
        """
        extracts all potential table names from input single or collection `TokenList`.

        :param Union[list[TokenList], TokenList] token_stream: a single `TokenList` or
                                                               collection of `TokenList`
                                                               objects to be checked.

        :rtype: list[str]
        """

        result = []
        if token_stream is not None:
            if not isinstance(token_stream, LIST_TYPES):
                token_stream = [token_stream]
            for item in token_stream:
                if item is not None:
                    if isinstance(item, IdentifierList):
                        for identifier in item.get_identifiers():
                            if isinstance(identifier, TokenList):
                                if identifier.get_real_name().lower() not in result:
                                    result.append(identifier.get_real_name().lower())
                            else:
                                if identifier.value.lower() not in result:
                                    result.append(identifier.value.lower())
                    elif isinstance(item, Identifier):
                        if item.get_real_name().lower() not in result:
                            result.append(item.get_real_name().lower())
                    elif item.ttype is Keyword:
                        if item.value not in result:
                            result.append(item.value.lower())
                    elif isinstance(item, Function):
                        if item.value not in result:
                            result.append(item.value.lower())

        return result

    def _extract_tables(self, sql, **options):
        """
        extracts all tables from given sql string expression.
        tables will be ordered as they appeared in input expression.

        :param str sql: sql expression be extracted for table names.

        :keyword bool include_select: specifies that select statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_insert: specifies that insert statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_update: specifies that update statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_delete: specifies that delete statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :returns list[str]
        :rtype: list
        """

        include_select = options.get('include_select', True)
        include_insert = options.get('include_insert', True)
        include_update = options.get('include_update', True)
        include_delete = options.get('include_delete', True)

        extracted_tables = {}
        statements = sqlparse.parse(sql)
        for statement in statements:
            stream = None
            if statement.get_type() != 'UNKNOWN':
                if include_select is True and self._is_subselect(statement):
                    stream = self._extract_select_from_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))
                if include_insert is True and self._is_insert(statement):
                    stream = self._extract_insert_into_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))
                if include_update is True and self._is_update(statement):
                    stream = self._extract_update_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))
                if include_delete is True and self._is_delete(statement):
                    stream = self._extract_delete_from_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))

        return list(extracted_tables.keys())

    def _get_identifier(self, token, force=False):
        """
        gets the identifier token available in given tokens children.

        :param Union[Token, TokenList] token: token to get its child identifier.

        :param bool force: specifies that if its child is not an
                           `Identifier` but it is a keyword, return it.
                           it could be helpful for some situations which
                           table names are equal to some keywords.

        :rtype: Union[Identifier, Keyword]
        """

        if token is not None:
            if isinstance(token, (Identifier, IdentifierList)):
                return token

            elif isinstance(token, Function) and force is True:
                return token.tokens[0]

            elif token.ttype is Keyword and force is True:
                return token

        return None



# -*- coding: utf-8 -*-
"""
start module.
"""

from APPLICATION_PACKAGE import APPLICATION_CLASS


app = APPLICATION_CLASS()

# the if condition is to ensure that multiprocessing
# on windows works as expected.
if __name__ == '__main__':
    app.run(use_reloader=False)



# -*- coding: utf-8 -*-
"""
authentication component module.
"""

from pyrin.application.decorators import component
from pyrin.application.context import Component

from APPLICATION_PACKAGE.security.authentication import AuthenticationPackage
from APPLICATION_PACKAGE.security.authentication.manager import AuthenticationManager


@component(AuthenticationPackage.COMPONENT_NAME, replace=True)
class AuthenticationComponent(Component, AuthenticationManager):
    """
    authentication component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
database orm query exceptions module.
"""

from pyrin.core.exceptions import CoreException


class DatabaseORMQueryException(CoreException):
    """
    database orm query exception.
    """
    pass


class DatabaseORMQueryBusinessException(DatabaseORMQueryException):
    """
    database orm query business exception.
    """
    pass


class ColumnsOutOfScopeError(DatabaseORMQueryBusinessException):
    """
    columns out of scope error.
    """
    pass


class UnsupportedQueryStyleError(DatabaseORMQueryException):
    """
    unsupported query style error.
    """
    pass



# -*- coding: utf-8 -*-
"""
packaging services module.
"""

from pyrin.packaging import PackagingPackage
from pyrin.application.services import get_component


def load_components(**options):
    """
    loads required packages and modules for application startup.
    """

    get_component(PackagingPackage.COMPONENT_NAME).load_components(**options)


def register_hook(instance):
    """
    registers the given instance into packaging hooks.

    :param PackagingHookBase instance: packaging hook instance to be registered.

    :raises InvalidPackagingHookTypeError: invalid packaging hook type error.
    """

    get_component(PackagingPackage.COMPONENT_NAME).register_hook(instance)



# -*- coding: utf-8 -*-
"""
permission manager module.
"""

from pyrin.security.permission.manager import PermissionManager as BasePermissionManager
from pyrin.core.exceptions import CoreNotImplementedError


class PermissionManager(BasePermissionManager):
    """
    permission manager class.
    """

    def _exists(self, *primary_key):
        """
        gets a value indicating that given permission exists in database.

        the input parameters could be as many as needed arguments to
        represent the primary key of your permission entity.
        if you don't want to use permissions concept in your application,
        you could leave this method unimplemented.

        :param object primary_key: permission primary key value.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
database orm sql extractor component module.
"""

from pyrin.application.decorators import component
from pyrin.application.context import Component
from pyrin.database.orm.sql.extractor import DatabaseORMSQLExtractorPackage
from pyrin.database.orm.sql.extractor.manager import DatabaseORMSQLExtractorManager


@component(DatabaseORMSQLExtractorPackage.COMPONENT_NAME)
class DatabaseORMSQLExtractorComponent(Component, DatabaseORMSQLExtractorManager):
    """
    database orm sql extractor component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
deserializer interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.context import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.utils.singleton import MultiSingletonMeta


class DeserializerSingletonMeta(MultiSingletonMeta):
    """
    deserializer singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractDeserializerBase(CoreObject, metaclass=DeserializerSingletonMeta):
    """
    abstract deserializer base class.
    """

    @abstractmethod
    def deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param object value: value to be deserialized.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: deserialized value.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def set_next(self, deserializer):
        """
        sets the next deserializer handler and returns it.

        :param Union[AbstractDeserializerBase, None] deserializer: deserializer instance to
                                                                   be set as next handler.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: AbstractDeserializerBase
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_accepted_type(self):
        """
        gets the accepted type for this deserializer
        which could deserialize values from this type.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
serializer keyed_tuple module.
"""

from pyrin.converters.serializer.base import SerializerBase
from pyrin.utils.sqlalchemy import keyed_tuple_to_dict


class CoreKeyedTupleSerializer(SerializerBase):
    """
    core keyed tuple serializer class.
    """

    def serialize(self, value, **options):
        """
        serializes the given value.

        :param AbstractKeyedTuple value: keyed tuple value to be serialized.

        :rtype: dict
        """

        return keyed_tuple_to_dict(value)



# -*- coding: utf-8 -*-
"""
migrations package.
"""

from pyrin.packaging.base import Package


class MigrationsPackage(Package):
    """
    migrations package class.
    """

    NAME = __name__
    ENABLED = False



# -*- coding: utf-8 -*-
"""
route base handler module.
"""

from werkzeug.routing import Rule

import pyrin.configuration.services as config_services
import pyrin.security.session.services as session_services

from pyrin.core.globals import _, LIST_TYPES
from pyrin.api.router.handlers.exceptions import InvalidViewFunctionTypeError, \
    MaxContentLengthLimitMismatchError, LargeContentError


class RouteBase(Rule):
    """
    route base class.
    """

    def __init__(self, rule, **options):
        """
        initializes a new instance of RouteBase.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls will be overwritten
                         if `replace=True` option is provided, otherwise an error
                         will be raised.

        :keyword tuple(str) methods: http methods that this route could handle.
                                     if not provided, defaults to `GET`, `HEAD`
                                     and `OPTIONS`.
                        
        :keyword callable view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule. pyrin
                               itself assumes the rule as endpoint if not provided.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint.
                                this is a bit tricky but useful if you want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the rule
                                only matches for the `default_subdomain` of the map. if the map is
                                not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword Union[string, Callable] redirect_to: if given this must be either a string
                                                      or callable. in case of a callable it's
                                                      called with the url adapter that
                                                      triggered the match and the values
                                                      of the url as keyword arguments and has
                                                      to return the target for the redirect,
                                                      otherwise it has to be a string with
                                                      placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        """

        methods = options.get('methods', ())
        if not isinstance(methods, LIST_TYPES):
            methods = (methods,)
            options.update(methods=methods)

        # we should call super method with exact param names because it
        # does not have `**options` in it's signature and raises an error
        # if extra keywords passed to it. maybe flask fixes it in the future.
        super().__init__(rule,
                         defaults=options.get('defaults', None),
                         subdomain=options.get('subdomain', None),
                         methods=methods,
                         build_only=options.get('build_only', False),
                         endpoint=options.get('endpoint', None),
                         strict_slashes=options.get('strict_slashes', None),
                         redirect_to=options.get('redirect_to', None),
                         alias=options.get('alias', False),
                         host=options.get('host', None))

        self._view_function = options.get('view_function')

        global_limit = config_services.get('api', 'general', 'max_content_length')

        restricted_length = options.get('max_content_length',
                                        config_services.get('api',
                                                            'general',
                                                            'restricted_max_content_length'))

        if restricted_length > global_limit:
            raise MaxContentLengthLimitMismatchError('Specified max content length '
                                                     '[{restricted}] for route [{route}] is '
                                                     'greater than global limit which is '
                                                     '[{global_limit}].'
                                                     .format(restricted=restricted_length,
                                                             route=rule,
                                                             global_limit=global_limit))

        self._max_content_length = restricted_length

        if not callable(self._view_function):
            raise InvalidViewFunctionTypeError('The provided view function [{function}] '
                                               'for route [{route}] is not callable.'
                                               .format(function=str(self._view_function),
                                                       route=str(self)))

    def handle(self, inputs, **options):
        """
        handles the current route.

        :param dict inputs: view function inputs.

        :raises LargeContentError: large content error.

        :returns: view function's result.

        :rtype: object
        """

        self._validate_content_length()
        self._handle(inputs, **options)

        return self._call_view_function(inputs, **options)

    def _handle(self, inputs, **options):
        """
        handles the current route.
        routes which need to perform extra operations before
        view function execution, must override this method.

        :param dict inputs: view function inputs.
        """
        pass

    def _validate_content_length(self):
        """
        validates content length for this route.

        :raises LargeContentError: large content error.
        """

        client_request = session_services.get_current_request()
        if client_request.safe_content_length > self.get_max_content_length():
            raise LargeContentError(_('Request content is too large.'))

    def _call_view_function(self, inputs, **options):
        """
        calls the route's view function.

        :param dict inputs: view function inputs.

        :returns: view function's result.

        :rtype: object
        """

        return self._view_function(**inputs)

    def get_view_function(self):
        """
        gets this route's view function.

        :rtype: callable
        """

        return self._view_function

    def get_max_content_length(self):
        """
        gets this route's max content bytes length.

        :rtype: int
        """

        return self._max_content_length



# -*- coding: utf-8 -*-
"""
model base module.
"""

import inspect

from sqlalchemy import inspect as sqla_inspect
from sqlalchemy.ext.declarative import as_declarative

import pyrin.database.services as database_services

from pyrin.core.context import CoreObject, DTO
from pyrin.database.model.exceptions import ColumnNotExistedError, EntityNotHashableError


@as_declarative(constructor=None)
class CoreEntity(CoreObject):
    """
    core entity class.
    it will be used as a base class for all models.
    """

    # holds the table name in database.
    __tablename__ = None

    # holds the extra arguments for table.
    # for example:
    # __table_args__ = {'schema': 'database_name.schema_name',
    #                   'extend_existing': True}
    __table_args__ = None

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreDeclarative.
        note that this method will only be called on user code, meaning
        that results returned by orm from database will not call `__init__`
        of each entity.

        it also sets attributes on the constructed instance using the
        names and values in `kwargs`. note that only keys that are present
        as mapped columns of the instance's class are allowed.

        :raises ColumnNotExistedError: column not existed error.
        """

        super().__init__()

        self._set_name(self.__class__.__name__)
        self.from_dict(False, **kwargs)

    def __eq__(self, other):
        if isinstance(other, self._get_root_base_class()):
            if self._is_primary_key_comparable(self.primary_key()) is True:
                return self.primary_key() == other.primary_key()
            else:
                return self is other

        return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        if self._is_primary_key_comparable(self.primary_key()) is False:
            raise EntityNotHashableError('Entity [{entity}] does not have '
                                         'a primary key so it is not hashable.'
                                         .format(entity=self))

        return hash('{root_base}.{pk}'.format(root_base=self._get_root_base_class(),
                                              pk=self.primary_key()))

    def __repr__(self):
        """
        gets the string representation of current entity.

        :rtype: str
        """
        return '{module}.{name} [{pk}]'.format(module=self.__module__,
                                               name=self.get_name(),
                                               pk=self.primary_key())

    def _is_primary_key_comparable(self, primary_key):
        """
        gets a value indicating that given primary key is comparable.
        the primary key is comparable if it is not None for single
        primary keys and if all the values in primary key tuple are
        not None for composite primary keys.

        :param Union[object, tuple[object]] primary_key: primary key value.

        :rtype: bool
        """

        if primary_key is None:
            return False

        if isinstance(primary_key, tuple):
            if len(primary_key) <= 0:
                return False
            else:
                return all(pk is not None for pk in primary_key)

        return True

    def primary_key(self, as_tuple=False):
        """
        gets the primary key value for this entity.
        it could be a single value or a tuple of values
        for composite primary keys.
        it could return None if no primary key is set for this entity.

        :param bool as_tuple: specifies that primary key value must be returned
                              as a tuple even if it's a single value.
                              defaults to False if not provided.

        :rtype: Union[object, tuple[object]]
        """

        columns = self.primary_key_columns()
        if len(columns) <= 0:
            return None

        if as_tuple is False and len(columns) == 1:
            return getattr(self, columns[0])
        else:
            return tuple(getattr(self, col) for col in columns)

    def _get_root_base_class(self):
        """
        gets root base class of this entity and caches it.
        root base class is the class which is direct subclass
        of CoreEntity in inheritance hierarchy.
        for example: {Base -> CoreEntity, A -> Base, B -> A}
        root base class of A, B and Base is Base class.

        :rtype: CoreEntity
        """

        base = getattr(self, '_root_base_class', None)
        if base is None:
            bases = inspect.getmro(type(self))
            base_entity_index = bases.index(CoreEntity) - 1
            base = bases[base_entity_index]
            self.__class__._root_base_class = base

        return base

    def _set_all_columns(self, columns):
        """
        sets all column names attribute for this class.

        :param tuple[str] columns: column names.
        """

        if getattr(self, '_all_columns', None) is None:
            self.__class__._all_columns = DTO()

        self.__class__._all_columns[type(self)] = columns

    def _set_exposed_columns(self, columns):
        """
        sets exposed column names attribute for this class.

        :param tuple[str] columns: column names.
        """

        if getattr(self, '_exposed_columns', None) is None:
            self.__class__._exposed_columns = DTO()

        self.__class__._exposed_columns[type(self)] = columns

    def _set_primary_key_columns(self, columns):
        """
        sets primary key column names attribute for this class.

        :param tuple[str] columns: column names.
        """

        if getattr(self, '_primary_key_columns', None) is None:
            self.__class__._primary_key_columns = DTO()

        self.__class__._primary_key_columns[type(self)] = columns

    def save(self):
        """
        saves the current entity.
        """

        database_services.get_current_store().add(self)
        return self

    def update(self, **kwargs):
        """
        updates the current entity with given values.
        """

        self.from_dict(**kwargs)
        return self.save()

    def delete(self):
        """
        deletes the current entity.
        """

        database_services.get_current_store().delete(self)

    def _get_primary_keys(self):
        """
        gets current entity's primary key columns if available.

        :rtype: tuple[str]
        """

        columns = getattr(self, '_primary_key_columns', None)
        if columns is not None:
            return columns.get(type(self), None)

        return None

    def primary_key_columns(self):
        """
        gets the primary key column name(s) of this entity.
        column names will be calculated once and cached.

        :returns: tuple[str]
        :rtype: tuple
        """

        pk = self._get_primary_keys()
        if pk is None:
            info = sqla_inspect(type(self))
            pk = tuple(info.get_property_by_column(col).key for col in info.primary_key)
            self._set_primary_key_columns(pk)

        return pk

    def all_columns(self):
        """
        gets all column names of entity.
        column names will be calculated once and cached.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = self._get_all_columns()
        if columns is None:
            info = sqla_inspect(type(self))
            columns = tuple(attr.key for attr in info.column_attrs)
            self._set_all_columns(columns)

        return columns

    def exposed_columns(self):
        """
        gets exposed column names of entity, which
        are those that have `exposed=True`.
        column names will be calculated once and cached.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = self._get_exposed_columns()
        if columns is None:
            info = sqla_inspect(type(self))
            columns = tuple(attr.key for attr in info.column_attrs
                            if attr.columns[0].exposed is True)
            self._set_exposed_columns(columns)

        return columns

    def _get_all_columns(self):
        """
        gets all column names of entity.
        returns None if not found.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = getattr(self, '_all_columns', None)
        if columns is not None:
            return columns.get(type(self), None)

        return None

    def _get_exposed_columns(self):
        """
        gets exposed column names of entity, which
        are those that have `exposed=True`.
        returns None if not found.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = getattr(self, '_exposed_columns', None)
        if columns is not None:
            return columns.get(type(self), None)

        return None

    def to_dict(self, exposed_only=True):
        """
        converts the entity into a dict and returns it.
        the result dict only contains the exposed columns of
        the entity which are those that their `exposed` attribute
        is set to True.

        :param bool exposed_only: if set to False, it returns all
                                  columns of the entity as dict.
                                  if not provided, defaults to True.

        :rtype: dict
        """

        columns_collection = self.exposed_columns
        if exposed_only is False:
            columns_collection = self.all_columns

        result = DTO()
        for col in columns_collection():
            result[col] = getattr(self, col)

        return result

    def from_dict(self, silent_on_invalid_column=True, **kwargs):
        """
        updates the column values of the entity from those
        values that are available in input keyword arguments.

        :keyword bool silent_on_invalid_column: specifies that if a key is not available
                                                in entity columns, do not raise an error.
                                                defaults to True if not provided.

        :raises ColumnNotExistedError: column not existed error.
        """

        all_columns = self.all_columns()
        for key, value in kwargs.items():
            if key in all_columns:
                setattr(self, key, value)
            else:
                if silent_on_invalid_column is False:
                    raise ColumnNotExistedError('Entity [{entity}] does not have '
                                                'a column attribute named [{column}].'
                                                .format(entity=self,
                                                        column=key))

    @classmethod
    def table_name(cls):
        """
        gets the table name that this entity represents in database.

        :rtype: str
        """

        return cls.__tablename__

    @classmethod
    def table_schema(cls):
        """
        gets the table schema that this entity represents in database.
        it might be `None` if schema has not been set for this entity.

        :rtype: str
        """

        schema = None
        if cls.__table_args__ is not None:
            schema = cls.__table_args__.get('schema', None)

        return schema

    @classmethod
    def table_fullname(cls):
        """
        gets the table fullname that this entity represents in database.
        fullname is `schema.table_name` if schema is available, otherwise it
        defaults to `table_name`.

        :rtype: str
        """

        schema = cls.table_schema()
        name = cls.table_name()

        if schema is not None:
            return '{schema}.{name}'.format(schema=schema, name=name)
        else:
            return name



# -*- coding: utf-8 -*-
"""
database manager module.
"""

from sqlalchemy import engine_from_config

import pyrin.configuration.services as config_services
import pyrin.logging.services as logging_services
import pyrin.security.session.services as session_services
import pyrin.utils.dictionary as dict_utils

from pyrin.core.globals import LIST_TYPES
from pyrin.database.hooks import DatabaseHookBase
from pyrin.core.mixin import HookMixin
from pyrin.database.model.base import CoreEntity
from pyrin.core.context import DTO, Manager
from pyrin.core.enumerations import ClientErrorResponseCodeEnum, ServerErrorResponseCodeEnum
from pyrin.database.interface import AbstractSessionFactoryBase
from pyrin.utils import response as response_utils
from pyrin.utils.custom_print import print_warning
from pyrin.database.exceptions import InvalidSessionFactoryTypeError, \
    DuplicatedSessionFactoryError, SessionFactoryNotExistedError, InvalidEntityTypeError, \
    InvalidDatabaseBindError


class DatabaseManager(Manager, HookMixin):
    """
    database manager class.
    """

    LOGGER = logging_services.get_logger('database')
    BIND_REMOVE_KEY_PREFIX = '__'
    DEFAULT_DATABASE_NAME = 'default'
    _hook_type = DatabaseHookBase

    def __init__(self):
        """
        initializes an instance of DatabaseManager.
        """

        super().__init__()

        # contains the application default database engine.
        self.___engine = self._create_default_engine()

        # a dictionary containing engines for different bounded databases.
        # in the form of: {str bind_name: Engine engine}
        self.__bounded_engines = self._create_bounded_engines()

        # a dictionary containing all entity classes that should be
        # bounded into a database engine other than the default one.
        # in the form of: {type entity: str bind_name}
        self._binds = DTO()

        # a dictionary containing entity to engine map for those entities that
        # should be bounded to a different database engine than the default one.
        # in the form of: {type entity: Engine engine}
        self._entity_to_engine_map = DTO()

        # a dictionary containing table name to engine map for those entities that
        # should be bounded to a different database engine than the default one.
        # in the form of: {string table_name: Engine engine}
        self._table_name_to_engine_map = DTO()

        # a dictionary containing session factories for request bounded and unbounded types.
        # in the for of: {bool request_bounded: Session session_factory}
        # it should have at most two different keys, True for request bounded
        # and False for request unbounded.
        self._session_factories = DTO()

    def get_current_store(self):
        """
        gets current database store.

        :returns: database session
        :rtype: Session
        """

        return self._get_current_session_factory()()

    def get_session_factory(self, request_bounded=None):
        """
        gets database session factory based on given input.

        this method should not be used directly for data manipulation.
        use `get_current_store` method instead.

        :param bool request_bounded: a value indicating that the session
                                     factory should be bounded into request.
                                     if not provided, it gets the current
                                     valid session factory.

        :returns: database session factory
        :rtype: Session
        """

        if request_bounded is None:
            return self._get_current_session_factory()

        return self._get_session_factory(request_bounded)

    def _get_current_session_factory(self):
        """
        gets database session factory that should be used.

        it simply checks that request is available in current
        context or not, and gets the correct session factory.

        :returns: database session factory
        :rtype: Session
        """

        return self._get_session_factory(session_services.is_request_context_available())

    def _get_session_factory(self, request_bounded):
        """
        gets the session factory based on given input.

        :param bool request_bounded: a value indicating that the session
                                     factory should be bounded into request.

        :raises SessionFactoryNotExistedError: session factory not existed error.

        :returns: database session factory
        :rtype: Session
        """

        if request_bounded not in self._session_factories.keys():
            raise SessionFactoryNotExistedError('Session factory with '
                                                'request_bounded={bounded} '
                                                'is not available.'
                                                .format(bounded=request_bounded))

        return self._session_factories.get(request_bounded)

    def _create_default_engine(self):
        """
        creates the default database engine using database
        configuration store and returns it.

        :returns: database engine
        :rtype: Engine
        """

        database_configs = config_services.get_active_section('database')
        return self._create_engine(database_configs)

    def _create_engine(self, database_configs, **kwargs):
        """
        creates a database engine using specified database configuration and returns it.

        each provided key from kwargs will override the corresponding
        key in database_configs dict. note that kwargs should not have any prefix.

        :param dict database_configs: database configs that should be used.

        :returns: database engine
        :rtype: Engine
        """

        configs_prefix = config_services.get_active('database', 'configs_prefix')
        return engine_from_config(database_configs, prefix=configs_prefix, **kwargs)

    def _create_bounded_engines(self):
        """
        creates all required bounded engines if any and returns them.

        :returns: dict(str bind_name: Engine engine)
        :rtype: dict
        """

        engines = DTO()
        base_database_configs = config_services.get_active_section('database')
        binds_names = base_database_configs['bind_names']

        if binds_names is None:
            return engines
        if not isinstance(binds_names, LIST_TYPES):
            binds_names = [binds_names]
        if len(binds_names) <= 0:
            return engines

        for name in binds_names:
            bind_configs = self._get_bind_configs(name)
            full_configs = self._merge_configs(base_database_configs, bind_configs)
            engine = self._create_engine(full_configs)
            engines[name] = engine

        return engines

    def _merge_configs(self, base_configs, bind_configs):
        """
        merges given base and bind configs and returns a new dict.

        merging will cause all keys in bind configs override any available key in
        base configs. also all keys that are present in bind configs and starting
        with double underscore `__` will be removed from result dict and also all
        keys that their names are exactly like prefixed ones but without `__` will
        be removed from result dict.

        :param dict base_configs: base configs from `database.config` file.
        :param dict bind_configs: bind configs from `database.binds.config`

        :rtype: dict
        """

        result = DTO(**base_configs)
        result.update(**bind_configs)
        result = dict_utils.remove_keys(result, self.BIND_REMOVE_KEY_PREFIX)

        return result

    def _get_bind_configs(self, bind_name):
        """
        gets all bind configs of given bind name for currently
        active environment in 'database.config' from 'database.binds.config' file.

        :param str bind_name: bind name to get its configs.

        :rtype: dict
        """

        bind_configs_section = self.get_bind_config_section_name(bind_name)
        return config_services.get_section('database.binds', bind_configs_section)

    def get_bind_config_section_name(self, bind_name):
        """
        gets the bind config section name for given bind name and currently
        active environment in 'database.config' from 'database.binds.config' file.

        :param str bind_name: bind name to get its section name.

        :rtype: str
        """

        active_environment = config_services.get_active_section_name('database')
        bind_config_key = '{environment}_{bind_name}'.format(environment=active_environment,
                                                             bind_name=bind_name)

        return bind_config_key

    def finalize_transaction(self, response):
        """
        this method will finalize database transaction of each request.

        we should not raise any exception in request handlers, so we return
        an error response in case of any exception.
        note that normally you should never call this method manually.

        :param CoreResponse response: response object.

        :rtype: CoreResponse
        """

        try:
            store = self.get_current_store()
            session_factory = self.get_session_factory()
            try:
                if response.status_code >= ClientErrorResponseCodeEnum.BAD_REQUEST:
                    store.rollback()
                    return response

                store.commit()
                return response
            except Exception:
                store.rollback()
                raise
            finally:
                session_factory.remove()
        except Exception as error:
            self.LOGGER.exception(str(error))
            return response_utils.make_exception_response(error,
                                                          code=ServerErrorResponseCodeEnum.
                                                          INTERNAL_SERVER_ERROR)

    def cleanup_session(self, exception):
        """
        this method will cleanup database session of each request.

        in case of any unhandled exception. we should not raise any exception
        in teardown request handlers, so we just log the exception.
        note that normally you should never call this method manually.

        :param Exception exception: exception instance.
        """

        if exception is not None:
            try:
                self.LOGGER.exception(str(exception))
                session_factory = self.get_session_factory()
                session_factory.remove()

            except Exception as error:
                self.LOGGER.exception(str(error))

    def register_session_factory(self, instance, **options):
        """
        registers a new session factory or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's is_request_bounded() is already available
        in registered session factories.

        :param AbstractSessionFactoryBase instance: session factory to be registered.
                                                    it must be an instance of
                                                    AbstractSessionFactoryBase.

        :keyword bool replace: specifies that if there is another registered
                               session factory with the same is_request_bounded(),
                               replace it with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidSessionFactoryTypeError: invalid session factory type error.
        :raises DuplicatedSessionFactoryError: duplicated session factory error.
        """

        if not isinstance(instance, AbstractSessionFactoryBase):
            raise InvalidSessionFactoryTypeError('Input parameter [{instance}] is '
                                                 'not an instance of [{base}].'
                                                 .format(instance=str(instance),
                                                         base=AbstractSessionFactoryBase))

        # checking whether is there any registered
        # instance with the same 'is_request_bounded()' value.
        if instance.is_request_bounded() in self._session_factories:
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicatedSessionFactoryError('There is another registered session factory '
                                                    'with "is_request_bounded={bounded}" but '
                                                    '"replace" option is not set, so session '
                                                    'factory [{instance}] could not be registered.'
                                                    .format(bounded=instance.is_request_bounded(),
                                                            instance=str(instance)))

            old_instance = self._session_factories[instance.is_request_bounded()]
            print_warning('Session factory [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=str(old_instance), new_instance=str(instance)))

        # registering new session factory.
        self._session_factories[instance.is_request_bounded()] = \
            instance.create_session_factory(self.get_default_engine())

    def register_bind(self, entity, bind_name, **options):
        """
        binds the given model class with specified bind database.

        :param CoreEntity entity: CoreEntity subclass to be bounded.
        :param str bind_name: bind name to be associated with the model class.

        :raises InvalidEntityTypeError: invalid entity type error.
        """

        if not issubclass(entity, CoreEntity):
            raise InvalidEntityTypeError('Input parameter [{entity}] is '
                                         'not a subclass of [{base}].'
                                         .format(entity=entity,
                                                 base=CoreEntity))

        # registering model into database binds.
        self._binds[entity] = bind_name

    def _map_entity_to_engine(self):
        """
        maps all application entities that should be bounded to a
        database engine other than the default one to relevant engines.

        :raises InvalidDatabaseBindError: invalid database bind error.
        """

        binds = config_services.get_active('database', 'bind_names')
        if binds is None:
            binds = []
        if not isinstance(binds, LIST_TYPES):
            binds = [binds]

        for entity, bind_name in self._binds.items():
            if bind_name not in binds:
                raise InvalidDatabaseBindError('Database bind name [{bind_name}] for entity '
                                               '[{entity_name}] is not available in '
                                               'database config store.'
                                               .format(bind_name=bind_name,
                                                       entity_name=entity))

            self._entity_to_engine_map[entity] = self.get_bounded_engines()[bind_name]
            self._table_name_to_engine_map[entity.table_name().lower()] = \
                self.get_bounded_engines()[bind_name]
            self._table_name_to_engine_map[entity.table_fullname().lower()] = \
                self.get_bounded_engines()[bind_name]

    def configure_session_factories(self):
        """
        configures all application session factories.

        normally, you should not call this method manually.

        :raises InvalidDatabaseBindError: invalid database bind error.
        """

        self._map_entity_to_engine()

        if len(self.get_entity_to_engine_map()) > 0:
            for key in self._session_factories:
                self._session_factories[key].configure(binds=self.get_entity_to_engine_map())

        self._after_session_factories_configured()

    def get_default_engine(self):
        """
        gets database default engine.

        :rtype: Engine
        """

        return self.___engine

    def get_bounded_engines(self):
        """
        gets database bounded engines.

        :returns: dict(str bind_name: Engine engine)
        :rtype: dict
        """

        return self.__bounded_engines

    def get_entity_to_engine_map(self):
        """
        gets entity to engine map.

        :returns: dict(type entity, Engine engine)
        :rtype: dict
        """

        return self._entity_to_engine_map

    def get_table_name_to_engine_map(self):
        """
        gets table name to engine map.

        :returns: dict(str table_name, Engine engine)
        :rtype: dict
        """

        return self._table_name_to_engine_map

    def _after_session_factories_configured(self):
        """
        this method will call `after_session_factories_configured`
        method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_session_factories_configured()

    def get_entity_engine(self, entity):
        """
        gets the database engine which the provided entity class is bounded to.

        :param CoreEntity entity: entity class to get its bounded engine.

        :rtype: Engine
        """

        if entity in self.get_entity_to_engine_map():
            return self.get_entity_to_engine_map()[entity]

        return self.get_default_engine()

    def get_table_engine(self, table_name):
        """
        gets the database engine which the provided table name is bounded to.

        :param str table_name: table name to get its bounded engine.

        :rtype: Engine
        """

        table_name = table_name.lower()
        if table_name in self.get_table_name_to_engine_map():
            return self.get_table_name_to_engine_map()[table_name]

        return self.get_default_engine()

    def get_bind_name_engine(self, bind_name):
        """
        gets the database engine which the provided bind name is bounded to.

        :param str bind_name: bind name to get its bounded engine.

        :raises InvalidDatabaseBindError: invalid database bind error.

        :rtype: Engine
        """

        if bind_name == self.DEFAULT_DATABASE_NAME:
            return self.get_default_engine()

        elif bind_name in self.get_bounded_engines():
            return self.get_bounded_engines()[bind_name]

        raise InvalidDatabaseBindError('Database bind name [{bind_name}] does '
                                       'not exist in database config store.'
                                       .format(bind_name=bind_name))

    def get_default_database_name(self):
        """
        gets default database name.

        :rtype: str
        """

        return self.DEFAULT_DATABASE_NAME



# -*- coding: utf-8 -*-
"""
router decorators module.
"""

import pyrin.api.router.services as router_services


def api(url, **options):
    """
    decorator to register an api handler for application.

    :param str url: url rule for this api.

    :keyword tuple(str) methods: http methods that this rule should handle.
                                 if not provided, defaults to `GET`.

    :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                to access this route's resource.

    :keyword bool login_required: specifies that this route could not be accessed
                                  if the requester has not a valid token.
                                  defaults to True if not provided.

    :keyword bool fresh_token: specifies that this route could not be accessed
                               if the requester has not a valid fresh token.
                               fresh token means a token that has been created by
                               providing user credentials to server.
                               defaults to False if not provided.

    :keyword bool replace: specifies that this route must replace
                           any existing route with the same url or raise
                           an error if not provided. defaults to False.

    :keyword int max_content_length: max content length that this route could handle,
                                     in bytes. if not provided, it will be set to
                                     `restricted_max_content_length` api config key.
                                     note that this value should be lesser than or equal
                                     to `max_content_length` api config key, otherwise
                                     it will cause an error.

    :raises DuplicateRouteURLError: duplicate route url error.

    :rtype: callable
    """

    def decorator(func):
        """
        decorates the given function and registers it as an api handler.

        :param callable func: function to register it as an api handler.

        :rtype: callable
        """

        router_services.add_route(url, view_func=func, **options)

        return func

    return decorator



# -*- coding: utf-8 -*-
"""
database paging services module.
"""

from pyrin.application.services import get_component
from pyrin.database.paging import DatabasePagingPackage


def get_default_limit():
    """
    gets default limit from database config store.

    :rtype: int
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).get_default_limit()


def is_auto_paging():
    """
    gets a value indicating that auto paging is enabled.

    :rtype: bool
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).is_auto_paging()


def enable_limit(options, limit=None):
    """
    enables limit with default or given limit parameter in given dict.

    :param int limit: custom limit value to be injected into given dict.
                      if not provided, defaults to limit value in
                      database config store.

    :param dict options: options dict to inject limit in it.
    """

    get_component(DatabasePagingPackage.COMPONENT_NAME).enable_limit(options, limit)


def disable_limit(options):
    """
    disables limit in given dict. it removes limit key from it.

    :param dict options: options dict to disable limit in it.
    """

    get_component(DatabasePagingPackage.COMPONENT_NAME).disable_limit(options)


def extract_limit(options):
    """
    extracts and gets limit parameter from given dict.
    it returns None if no limit is set in given dict.

    :param dict options: options dict to get limit from it.

    :rtype: int
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).extract_limit(options)


def extract_offset(options):
    """
    extracts and gets offset parameter from given dict.
    it returns None if no offset is set in given dict.

    :param dict options: options dict to get offset from it.

    :rtype: int
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).extract_offset(options)


def enable_paging(options, offset, limit=None):
    """
    enables paging in given dict.
    it adds a paging key in given dict and sets its value to
    True. it also injects default limit and offset into it

    :param dict options: options dict to enable paging in it.
    :param int offset: offset value to be set in given dict.
    :param int limit: limit value to be set in given dict.
    """

    get_component(DatabasePagingPackage.COMPONENT_NAME).enable_paging(options, offset, limit)


def disable_paging(options):
    """
    disables paging in given dict.
    it removes paging key in given dict and also
    removes keys for limit and offset.

    :param dict options: options dict to disable paging in it.
    """

    get_component(DatabasePagingPackage.COMPONENT_NAME).disable_paging(options)


def extract_paging(options):
    """
    extracts and gets limit and offset parameters from given dict.
    if limit or offset key is not available in provided
    dict, it gets the value from database config store
    from `paging` section for the absent key.

    :param dict options: options dict to get limit and offset from it.

    :returns: tuple(int limit, int offset)
    :rtype: tuple
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).extract_paging(options)


def is_paging_required(options):
    """
    gets a value indicating that paging should be done.
    it checks that paging key name is available in dict
    and its value is set to True, otherwise returns False.

    :param dict options: options dict to detect paging should be done from it.

    :rtype: bool
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).is_paging_required(options)



# -*- coding: utf-8 -*-
"""
security manager module.
"""

from pyrin.security.manager import SecurityManager as BaseSecurityManager
from pyrin.core.exceptions import CoreNotImplementedError


class SecurityManager(BaseSecurityManager):
    """
    security manager class.
    """

    def has_permission(self, user, permissions, **options):
        """
        gets a value indicating that given user has the specified permissions.

        :param user: user identity to check its permissions.
        :param list[Permission] permissions: permissions to check for user.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
permission base module.
"""

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.permission.base import PermissionBase


class Permission(PermissionBase):
    """
    permission class.

    all application permissions must be an instance of this.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of Permission.

        input parameters of this method must be customized
        based on your application's design requirements.
        """

        # set the required attributes and check validations here, above the `super()` call.

        super().__init__(*args, **options)

    def __str__(self):
        """
        gets the string representation of current permission.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def to_entity(self):
        """
        gets the equivalent entity of current permission.

        :rtype: CoreEntity
        """

        raise CoreNotImplementedError()

    def get_id(self):
        """
        gets permission id.

        it could return a single value or a combination of multiple values
        (ex. a tuple). note that the returned value must be fully unique for
        each different permission and also it must be a hashable value to
        be used as dict key.

        :rtype: object
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
utils file module.
"""

import os

import pyrin.utils.path as path_utils

from pyrin.utils.exceptions import IsNotDirectoryError


def replace_file_values(source, data):
    """
    replaces the values in given file with values available in given dict.

    the file must contain dict keys to be replaced by dict values for that keys.

    :param str source: file path to replace its values.
                       it must be an absolute path.

    :param dict data: a dict containing all values that
                      must be replaced in given file.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    path_utils.assert_exists(source)

    if data is not None and len(data) > 0:
        with open(source, 'r') as file:
            file_data = file.read()

        for key, value in data.items():
            file_data = file_data.replace(key, value)

        with open(source, 'w') as file:
            file.write(file_data)


def replace_files_values(source, data, *patterns):
    """
    replaces the values in all files available in given
    directory with values available in given dict.
    if patterns is given, only matching files will be included.

    :param str source: directory path to process its files.
                       the operation will also include all
                       files of all subdirectories.
                       it must be an absolute path.

    :param dict data: a dict containing all values that
                      must be replaced in given file.

    :param str patterns: file name end pattern to include in
                         replace operation. for example it
                         could be `'.py', '.html'`.
                         all files will be included on if not provided.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    :raises IsNotDirectoryError: is not directory error.
    """

    path_utils.assert_exists(source)

    if not os.path.isdir(source):
        raise IsNotDirectoryError('Provided path [{source}] is not a directory.'
                                  .format(source=source))

    for path, directories, files in os.walk(source, followlinks=True):
        for name in files:
            if is_match(name, *patterns):
                file_path = os.path.abspath(os.path.join(path, name))
                replace_file_values(file_path, data)


def is_match(source, *patterns):
    """
    gets a value indicating that given file name end, matches with any of given patterns.

    :param str source: source file path.

    :param str patterns: file name end pattern. for example it could
                         be `'.py', '.html'`. it will match all file
                         names if no pattern is provided.

    :rtype: bool
    """

    if len(patterns) <= 0:
        return True

    for pattern in patterns:
        if source.lower().endswith(pattern.lower()):
            return True

    return False



# -*- coding: utf-8 -*-
"""
deserializer string module.
"""

import re

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase
from pyrin.core.globals import NULL


@deserializer()
class StringDeserializer(StringPatternDeserializerBase):
    """
    string deserializer class.
    """

    # default min for this deserializer is 2 because it should
    # at least has two single or double quotes at both ends.
    DEFAULT_MIN = 2

    # matches a string value, all of these values will be matched.
    # example: "sample_value", "12345678", 'true'
    # note that the second and third examples will be deserialized as a
    # string with value 12345678 and true and won't deserialize as an
    # integer or boolean, because of the enclosing single or double quotes.
    DOUBLE_QUOTE_REGEX = re.compile(r'^[\"].*[\"]$')
    SINGLE_QUOTE_REGEX = re.compile(r'^[\'].*[\']$')

    def __init__(self, **options):
        """
        creates an instance of StringDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted patterns
                                                             and their length for string
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: str
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        # removing the first and last single or double quotes from value.
        value = value[1:-1]

        return value

    def get_default_formats(self):
        """
        gets default accepted patterns that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(self.DOUBLE_QUOTE_REGEX, self.UNDEF_LENGTH),
                (self.SINGLE_QUOTE_REGEX, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
public route handler module.
"""

from pyrin.api.router.handlers.base import RouteBase


class PublicRoute(RouteBase):
    """
    public route class.
    this class should be used to manage application public api
    routes that does not need valid token.
    """

    def __init__(self, rule, **options):
        """
        initializes a new instance of PublicRoute.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls will be overwritten
                         if `replace=True` option is provided, otherwise an error
                         will be raised.

        :keyword tuple(str) methods: http methods that this route could handle.
                                     if not provided, defaults to `GET`, `HEAD`
                                     and `OPTIONS`.

        :keyword callable view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule. pyrin
                               itself assumes the rule as endpoint if not provided.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint.
                                this is a bit tricky but useful if you want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the rule
                                only matches for the `default_subdomain` of the map. if the map is
                                not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only
                                      for this rule. if not specified the `Map` setting is used.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword Union[string, Callable] redirect_to: if given this must be either a string
                                                      or callable. in case of a callable it's
                                                      called with the url adapter that triggered
                                                      the match and the values of the url as
                                                      keyword arguments and has to return the
                                                      target for the redirect, otherwise it
                                                      has to be a string with placeholders
                                                      in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        """

        super().__init__(rule, **options)



# -*- coding: utf-8 -*-
"""
database orm package.
"""

from pyrin.packaging.base import Package


class DatabaseORMPackage(Package):
    """
    database orm package class.
    """

    NAME = __name__
    DEPENDS = []



# -*- coding: utf-8 -*-
"""
caching handlers memcached module.
"""

import pyrin.configuration.services as config_services

from pyrin.caching.handlers.base import CachingHandlerBase


class MemcachedCachingHandler(CachingHandlerBase):
    """
    memcached caching handler class.
    """

    def __init__(self, **options):
        """
        initializes an instance of MemcachedCachingHandler.
        """

        options.update(timeout=config_services.get('caching', 'memcached', 'timeout'))

        super().__init__('memcached', **options)



# -*- coding: utf-8 -*-
"""
database orm query base module.
"""

import inspect

from sqlalchemy import inspection, log, func
from sqlalchemy.orm import Query, lazyload
from sqlalchemy.orm.attributes import InstrumentedAttribute

from pyrin.core.globals import LIST_TYPES, _
from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.services import get_current_store
from pyrin.database.orm.query.exceptions import ColumnsOutOfScopeError, \
    UnsupportedQueryStyleError


@inspection._self_inspects
@log.class_logger
class CoreQuery(Query):
    """
    core query class.
    this is the application default query class.
    it extends sqlalchemy `Query` class.
    """

    def __init__(self, entities, session=None, **options):
        """
        initializes an instance of CoreQuery.
        this method has been overridden to provide the concept of scope to queries.
        it is useful if you want to let users (end users not developers) to select
        which columns they want to be returned in a service. in this situation, if
        there is no scope defined, they could add any columns of other entities, but
        using scope, prevents this. but on normal use cases there is no need to define
        scope, and its also more efficient.

        :param tuple entities: entities or columns that are needed for query.
        :param Session session: optional session object to bind this query to it.

        :keyword Union[type, tuple[type]] scope: class type of the entities that this
                                                 query instance will work on. if the
                                                 query is working on multiple entities,
                                                 this value must be a tuple of all class
                                                 types of that entities.

                                                 for example: if you set
                                                 `entities=SomeEntity.id, AnotherEntity.name`
                                                 you should leave `scope=None` to skip validation
                                                 or you could set
                                                 `scope=(SomeEntity, AnotherEntity)`
                                                 this way validation succeeds, but if
                                                 you set `scope=SomeEntity`
                                                 then the query will not be executed
                                                 and an error will be raised.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope = options.get('scope', None)
        if scope is not None:
            self._prepare_validation(entities, scope)

        super().__init__(entities, session)

    def _validate_scope(self, entities, scope):
        """
        validates the query for given entities based on provided scope.

        :param tuple entities: entities or columns that needed for query.

        :param Union[type, tuple[type]] scope: class type of the entities that this
                                               query instance will work on. if the
                                               query is working on multiple entities,
                                               this value must be a tuple of all class
                                               types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope_classes = set(entity for entity in scope if inspect.isclass(entity)
                            and issubclass(entity, CoreEntity))

        requested_classes = set(entity for entity in entities if inspect.isclass(entity)
                                and issubclass(entity, CoreEntity))

        requested_classes_by_column = set(column.class_ for column in entities if
                                          isinstance(column, InstrumentedAttribute)
                                          and issubclass(getattr(column, 'class_', type),
                                                         CoreEntity))

        all_requested_classes = requested_classes.union(requested_classes_by_column)

        if not all_requested_classes.issubset(scope_classes):
            raise ColumnsOutOfScopeError(_('Requested columns are out of scope of this '
                                           'query. Please revise the requested columns.'))

    def _prepare_validation(self, entities, scope):
        """
        prepares entities and scope and validates them if required.

        :param tuple entities: entities or columns that needed for query.

        :param Union[type, tuple[type]] scope: class type of the entities that this
                                               query instance will work on. if the
                                               query is working on multiple entities,
                                               this value must be a tuple of all class
                                               types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        if scope is None or entities is None:
            return

        if not isinstance(scope, LIST_TYPES):
            scope = (scope,)

        if not isinstance(entities, LIST_TYPES):
            entities = (entities,)

        if len(scope) <= 0 or len(entities) <= 0:
            return

        self._validate_scope(entities, scope)

    def count(self, **options):
        """
        returns the count of rows the sql formed by this `Query` would return.
        this method is overridden to prevent inefficient count() of sqlalchemy `Query`
        which produces a subquery.

        this method generates a single sql query like below:
        select count(column, ...)
        from table
        where ...

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.

        :keyword bool fallback: specifies that count should
                                be executed using original sqlalchemy
                                function which produces a subquery.
                                defaults to False if not provided.

        :rtype: int
        """

        fallback = options.get('fallback', False)
        needs_fallback = False
        columns = []
        # if there is group by clause, a subquery
        # is inevitable to be able to get count.
        if self.selectable._group_by_clause is not None and \
                self.selectable._group_by_clause.clauses is not None and \
                len(self.selectable._group_by_clause.clauses) > 0:
            needs_fallback = True
        else:
            for single_column in self.selectable.columns:
                if not isinstance(single_column, CoreColumn):
                    if fallback is False:
                        raise UnsupportedQueryStyleError('Current query does not have columns '
                                                         'of type [{column_type}] in its '
                                                         'expression. if you need to apply a '
                                                         '"DISTINCT" keyword, you should apply '
                                                         'it by passing "distinct=True" keyword '
                                                         'to count() method and do not apply it '
                                                         'in query structure itself. for example '
                                                         'instead of writing "store.query('
                                                         'distinct(Entity.id)).count()" you '
                                                         'should write this in the following '
                                                         'form "store.query(Entity.id).count('
                                                         'distinct=True)". but if you want the '
                                                         'sqlalchemy original style of count() '
                                                         'which produces a subquery, it is also '
                                                         'possible to fallback to that default '
                                                         'sqlalchemy count() but keep in mind '
                                                         'that, that method is not efficient. '
                                                         'you could pass "fallback=True" in '
                                                         'options to fallback to default mode '
                                                         'if overridden count() method failed '
                                                         'to provide count.'
                                                         .format(column_type=CoreColumn))
                    else:
                        needs_fallback = True
                        break

                fullname = single_column.fullname()
                if fullname not in (None, ''):
                    columns.append(fullname)

        if needs_fallback is True:
            return super().count()

        func_count = func.count()
        if len(columns) > 0:
            distinct = options.get('distinct', False)
            column_clause = ', '.join(columns)
            if distinct is True:
                column_clause = 'distinct {clause}'.format(clause=column_clause)
            func_count = func.count(column_clause)

        statement = self.options(lazyload('*')).statement.with_only_columns(
            [func_count]).order_by(None)

        store = get_current_store()
        return store.execute(statement).scalar()



# -*- coding: utf-8 -*-
"""
deserializer tuple module.
"""

import re

import pyrin.converters.deserializer.services as deserializer_services

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import DeserializerBase, \
    StringPatternDeserializerBase
from pyrin.core.globals import NULL


@deserializer()
class TupleDeserializer(DeserializerBase):
    """
    tuple deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of TupleDeserializer.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes every possible value available in input tuple.
        and gets a new deserialized tuple.

        :param tuple value: value that should be deserialized.

        :rtype: tuple
        """

        if not self.is_deserializable(value, **options):
            return NULL

        result_list = [item for item in value]

        index = 0
        for item in result_list:
            deserialized_value = NULL

            if self.is_deserializable(item, **options):
                deserialized_value = self.deserialize(item)
            else:
                deserialized_value = deserializer_services.deserialize(item, **options)

            if deserialized_value is not NULL:
                result_list[index] = deserialized_value

            index += 1
            continue

        return tuple(result_list)

    def get_accepted_type(self):
        """
        gets the accepted type for this deserializer
        which could deserialize values from this type.

        :rtype: type
        """

        return tuple


@deserializer()
class StringTupleDeserializer(StringPatternDeserializerBase):
    """
    string tuple deserializer class.
    note that this deserializer could only handle tuples with single depth.
    meaning that nested tuples are not supported. and also nested lists or
    dictionaries or sets or any other collections are not supported and
    stops deserialization.
    for example: (1, (2, 4), [5, 4]) will not be deserialized.
    """

    # default min for this deserializer is 2 because
    # it should at least has ( and ) at both ends.
    DEFAULT_MIN = 2

    # matches a tuple inside string, all of these values will be matched.
    # example: (), (1), (1,), (1,2), (1,2,)
    # it won't accept nested collections, all of these values won't match.
    # example: ([]), (1, {2: 4}, (2,3))
    TUPLE_REGEX = re.compile(r'^\(\)$|^\([^\(\){}\[\]]+(,[^\(\){}\[\]]+)*\)$')

    def __init__(self, **options):
        """
        creates an instance of StringTupleDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted patterns
                                                             and their length for tuple
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: tuple
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        value = value.strip()

        # removing the first '(' and last ')' from value.
        value = value[1:-1]
        items = value.split(',')
        temp_list = []
        for item in items:
            if len(item.strip()) > 0:
                temp_list.append(item.strip())

        # this deserializer does not handle nested tuples, so it won't
        # check whether each item is deserializable or not.
        return deserializer_services.deserialize(tuple(temp_list), **options)

    def get_default_formats(self):
        """
        gets default accepted patterns that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(self.TUPLE_REGEX, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
router services module.
"""

from pyrin.api.router import RouterPackage
from pyrin.application.decorators import route_factory
from pyrin.application.services import get_component


@route_factory()
def create_route(rule, **options):
    """
    creates the appropriate route based on the input parameters.

    :param str rule: unique url rule to register this route for.
                     routes with duplicated urls will be overwritten
                     if `replace=True` option is provided, otherwise an error
                     will be raised.

    :keyword tuple(str) methods: http methods that this route could handle.
                                 if not provided, defaults to `GET`, `HEAD`
                                 and `OPTIONS`.

    :keyword callable view_function: a function to be called on accessing this route.

    :keyword str endpoint: the endpoint for the registered url rule. pyrin
                           itself assumes the url rule as endpoint if not provided.

    :keyword bool login_required: specifies that this route could not be accessed
                                  if the requester has not a valid token.
                                  defaults to True if not provided.

    :keyword bool fresh_token: specifies that this route could not be accessed
                               if the requester has not a valid fresh token.
                               fresh token means a token that has been created by
                               providing user credentials to server.
                               defaults to False if not provided.

    :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                to access this route's resource.

    :keyword int max_content_length: max content length that this route could handle,
                                     in bytes. if not provided, it will be set to
                                     `restricted_max_content_length` api config key.
                                     note that this value should be lesser than or equal
                                     to `max_content_length` api config key, otherwise
                                     it will cause an error.

    :raises RouteAuthenticationMismatchError: route authentication mismatch error.

    :rtype: RouteBase
    """

    return get_component(RouterPackage.COMPONENT_NAME).create_route(rule, **options)


def add_route(url, endpoint=None, view_func=None,
              provide_automatic_options=None, **options):
    """
    connects a url rule. if a view_func is provided it will be registered with the endpoint.
    if there is another rule with the same url and `replace=True` option is provided,
    it will be replaced, otherwise an error will be raised.

    :param str url: the url rule as string.

    :param str endpoint: the endpoint for the registered url rule.
                         pyrin itself assumes the url rule as endpoint.

    :param callable view_func: the function to call when serving a request to the
                               provided endpoint.

    :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                           added automatically.
                                           this can also be controlled by setting the
                                           `view_func.provide_automatic_options = False`
                                           before adding the rule.

    :keyword tuple(str) methods: http methods that this rule should handle.
                                 if not provided, defaults to `GET`.

    :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                to access this route's resource.

    :keyword bool login_required: specifies that this route could not be accessed
                                  if the requester has not a valid token.
                                  defaults to True if not provided.

    :keyword bool fresh_token: specifies that this route could not be accessed
                               if the requester has not a valid fresh token.
                               fresh token means a token that has been created by
                               providing user credentials to server.
                               defaults to False if not provided.

    :keyword bool replace: specifies that this route must replace
                           any existing route with the same url or raise
                           an error if not provided. defaults to False.

    :keyword int max_content_length: max content length that this route could handle,
                                     in bytes. if not provided, it will be set to
                                     `restricted_max_content_length` api config key.
                                     note that this value should be lesser than or equal
                                     to `max_content_length` api config key, otherwise
                                     it will cause an error.

    :raises DuplicateRouteURLError: duplicate route url error.
    """

    return get_component(RouterPackage.COMPONENT_NAME).add_route(url, endpoint, view_func,
                                                                 provide_automatic_options,
                                                                 **options)



# -*- coding: utf-8 -*-
"""
alembic handlers params module.
"""

import pyrin.globalization.datetime.services as datetime_services

from pyrin.cli.arguments import BooleanArgument, KeywordArgument, PositionalArgument
from pyrin.database.migration.alembic.interface import AlembicCLIParamBase


class SQLParam(BooleanArgument, AlembicCLIParamBase):
    """
    sql param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SQLParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('sql', '--sql', default=default)


class TagParam(KeywordArgument, AlembicCLIParamBase):
    """
    tag param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of TagParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('tag', '--tag', default=default)


class RevisionParam(PositionalArgument, AlembicCLIParamBase):
    """
    revision param class.
    """

    def __init__(self, index=None, default=None):
        """
        initializes an instance of RevisionParam.

        :param int index: zero based index of this param in cli command inputs.
                          defaults to 0 if not provided.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        if index is None:
            index = 0

        super().__init__('revision', index, default=default)


class ResolveDependenciesParam(BooleanArgument, AlembicCLIParamBase):
    """
    resolve dependencies param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of ResolveDependenciesParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('resolve_dependencies', '--resolve-dependencies', default=default)


class IndicateCurrentParam(BooleanArgument, AlembicCLIParamBase):
    """
    indicate current param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of IndicateCurrentParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('indicate_current', '--indicate-current', default=default)


class RevisionRangeParam(KeywordArgument, AlembicCLIParamBase):
    """
    revision range param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of RevisionRangeParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('revision_range', '--rev-range', default=default)


class MessageParam(KeywordArgument, AlembicCLIParamBase):
    """
    message param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of MessageParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('message', '--message', default=default)

    def _process_inputs(self, **options):
        """
        processes the inputs in given dict.

        :rtype: dict
        """

        if options.get('message', None) is None:
            message = datetime_services.get_current_timestamp(date_sep=None,
                                                              main_sep=None,
                                                              time_sep=None)
            options.update(message=message)

        return super()._process_inputs(**options)


class BranchLabelParam(KeywordArgument, AlembicCLIParamBase):
    """
    branch label param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of BranchLabelParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('branch_label', '--branch-label', default=default)


class RevisionIDParam(KeywordArgument, AlembicCLIParamBase):
    """
    revision id param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of RevisionIDParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('revision_id', '--rev-id', default=default)


class RevisionsParam(PositionalArgument, AlembicCLIParamBase):
    """
    revisions param class.
    """

    def __init__(self, index=None, default=None):
        """
        initializes an instance of RevisionsParam.

        :param int index: zero based index of this param in cli command inputs.
                          defaults to 0 if not provided.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        if index is None:
            index = 0

        super().__init__('revisions', index, default=default)


class AutoGenerateParam(BooleanArgument, AlembicCLIParamBase):
    """
    autogenerate param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of AutoGenerateParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('autogenerate', '--autogenerate', default=default)


class HeadParam(KeywordArgument, AlembicCLIParamBase):
    """
    head param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of HeadParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('head', '--head', default=default)


class SpliceParam(BooleanArgument, AlembicCLIParamBase):
    """
    splice param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SpliceParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('splice', '--splice', default=default)


class VersionPathParam(KeywordArgument, AlembicCLIParamBase):
    """
    version path param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of VersionPathParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('version_path', '--version-path', default=default)


class DependsOnParam(KeywordArgument, AlembicCLIParamBase):
    """
    depends on param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of DependsOnParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('depends_on', '--depends-on', default=default)


class PurgeParam(BooleanArgument, AlembicCLIParamBase):
    """
    purge param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of PurgeParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to None if not provided.
        """

        super().__init__('purge', '--purge', default=default)



# -*- coding: utf-8 -*-
"""
manage module.

to enable locale management for application, execute:
`python manage.py babel enable`

to enable migrations for application, execute:
`python manage.py alembic enable`
"""

import fire

from pyrin.database.migration.alembic.cli import AlembicCLI
from pyrin.globalization.locale.babel.cli import BabelCLI
from pyrin.utils.environment import set_python_path

set_python_path('.')

from APPLICATION_PACKAGE import APPLICATION_CLASS


app_instance = APPLICATION_CLASS(scripting_mode=True)


class Groups(object):
    """
    groups class.

    each cli handler group is resided in its relevant group name.

    usage example:

    `python manage.py alembic upgrade --arg value`
    `python manage.py babel extract --arg value`
    """

    def __init__(self):
        """
        initializes an instance of Groups.
        """

        self.babel = BabelCLI()
        self.alembic = AlembicCLI()


if __name__ == '__main__':
    fire.Fire(Groups())



# -*- coding: utf-8 -*-
"""
logging decorators module.
"""

import time

from functools import update_wrapper

import pyrin.logging.services as logging_services
import pyrin.configuration.services as config_services


def audit(func):
    """
    decorator to log information before and after
    a function execution in debug mode.

    :param callable func: function.

    :returns: function result.
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and logs it's behavior in debug mode
        and returns the function's result every time that function gets called.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :returns: function result.
        """

        if config_services.get_active('logging', 'audit_log') is not True:
            return func(*args, **kwargs)

        start_time = time.time()
        try:
            return func(*args, **kwargs)

        finally:
            end_time = time.time()
            logging_services.debug('Duration of function call [{module}.{name}]: [{time} ms].'
                                   .format(module=func.__module__,
                                           name=func.__name__,
                                           time='{:0.5f}'
                                           .format((end_time - start_time) * 1000)))

    return update_wrapper(decorator, func)



# -*- coding: utf-8 -*-
"""
api manager module.
"""

import pyrin.logging.services as logging_services
import pyrin.configuration.services as config_services

from pyrin.core.context import Manager
from pyrin.core.enumerations import ServerErrorResponseCodeEnum
from pyrin.utils import response as response_utils
from pyrin.core.globals import _


class APIManager(Manager):
    """
    api manager class.
    """

    LOGGER = logging_services.get_logger('api')

    def handle_http_error(self, exception):
        """
        handles http exceptions.
        note that normally you should never call this method manually.

        :param HTTPException exception: exception instance.

        :rtype: CoreResponse
        """

        self._log_exception(exception)
        return response_utils.make_exception_response(exception)

    def handle_server_business_error(self, exception):
        """
        handles server internal core business exceptions.
        note that normally you should never call this method manually.

        :param CoreBusinessException exception: core business exception instance.

        :rtype: CoreResponse
        """

        self._log_exception(exception)
        return response_utils.make_exception_response(exception)

    def handle_server_error(self, exception):
        """
        handles server internal core exceptions.
        note that normally you should never call this method manually.
        in any environment which debug mode is False, the original error
        message will be replaced by a generic error message before being
        sent to client for security reasons.

        :param CoreException exception: core exception instance.

        :rtype: CoreResponse
        """

        self._log_exception(exception)
        if config_services.get_active('environment', 'debug') is True:
            return response_utils.make_exception_response(exception)

        return response_utils.make_error_response(self._get_generic_error_message(),
                                                  code=exception.code)

    def handle_server_unknown_error(self, exception):
        """
        handles unknown server internal exceptions.
        note that normally you should never call this method manually.
        in any environment which debug mode is False, the original error
        message will be replaced by a generic error message before being
        sent to client for security reasons.

        :param Exception exception: exception instance.

        :rtype: CoreResponse
        """

        self._log_exception(exception)
        if config_services.get_active('environment', 'debug') is True:
            return response_utils.make_exception_response(exception,
                                                          code=ServerErrorResponseCodeEnum.
                                                          INTERNAL_SERVER_ERROR)

        return response_utils.make_error_response(self._get_generic_error_message(),
                                                  code=ServerErrorResponseCodeEnum.
                                                  INTERNAL_SERVER_ERROR)

    def _get_generic_error_message(self):
        """
        gets generic error message to be sent to client
        for any environment which debug mode is False.

        :rtype: str
        """

        return _('Application has been encountered an error. Please '
                 'contact the support team if problem persists.')

    def _log_exception(self, exception):
        """
        logs the input exception.

        :param Exception exception: exception instance to be logged.
        """

        self.LOGGER.exception(str(exception))



# -*- coding: utf-8 -*-
"""
utils string module.
"""


def upper(value):
    """
    returns the uppercase copy of input string.

    :param str value: string to make uppercase.

    :rtype: str
    """

    return value.upper()


def lower(value):
    """
    returns the lowercase copy of input string.

    :param str value: string to make lowercase.

    :rtype: str
    """

    return value.lower()


def remove_line_break_escapes(value):
    """
    removes line break escapes from given value.
    it replaces `\\n` with `\n` to enable line breaks.

    :param str value: value to remove line break escapes from it.

    :rtype: str
    """

    return value.replace('\\n', '\n')


def remove_duplicate_space(value):
    """
    removes all duplicate spaces and keeps just a single space in given value.

    :param str value: value to remove duplicate spaces from it.

    :rtype: str
    """

    if value is None:
        return None

    while '  ' in value:
        value = value.replace('  ', ' ')

    return value



# -*- coding: utf-8 -*-
"""
permission hooks module.
"""

import pyrin.security.permission.services as permission_services

from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def __init__(self):
        """
        initializes an instance of ApplicationHook.
        """

        super().__init__()

    def before_application_start(self):
        """
        this method will be called before application gets started.
        note that this method will not get called when application
        starts in scripting mode.
        """

        # we should synchronize all application permissions with database.
        permission_services.synchronize_all()



# -*- coding: utf-8 -*-
"""
authentication manager module.
"""

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.authentication.manager import AuthenticationManager as \
    BaseAuthenticationManager


class AuthenticationManager(BaseAuthenticationManager):
    """
    authentication manager class.
    """

    def _push_custom_data(self, header, payload, **options):
        """
        pushes the custom data into current request from input values.

        :param dict header: token header data.

        :param dict payload: payload data of authenticated token.
        :type payload: dict(str type: token type)
        """

        raise CoreNotImplementedError()

    def _validate_custom(self, header, payload, **options):
        """
        validates the given inputs for custom attributes.
        an error will be raised if validation fails.

        :param dict header: token header data.

        :param dict payload: payload data to be validated.
        :type payload: dict(str type: token type)

        :raises AuthenticationFailedError: authentication failed error.
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
model services module.
"""

from pyrin.application.services import get_component
from pyrin.database.model import ModelPackage


def is_core_entity(value):
    """
    gets a value indicating that input value is an instance of `CoreEntity`.
    this method could be used where direct reference to `CoreEntity`
    produces an error.

    :param object value: value to be checked.

    :rtype: bool
    """

    return get_component(ModelPackage.COMPONENT_NAME).is_core_entity(value)


def is_abstract_keyed_tuple(value):
    """
    gets a value indicating that input value is an instance
    of `AbstractKeyedTuple`. `AbstractKeyedTuple` objects are
    those objects that are returned by sqlalchemy `Query`
    with columns or multiple entities.

    :param object value: value to be checked.

    :rtype: bool
    """

    return get_component(ModelPackage.COMPONENT_NAME).is_abstract_keyed_tuple(value)



# -*- coding: utf-8 -*-
"""
router manager module.
"""

import pyrin.application.services as application_services

from pyrin.api.router.exceptions import RouteAuthenticationMismatchError
from pyrin.api.router.handlers.protected import ProtectedRoute, FreshProtectedRoute
from pyrin.api.router.handlers.public import PublicRoute
from pyrin.core.context import Manager


class RouterManager(Manager):
    """
    router manager class.
    """

    def create_route(self, rule, **options):
        """
        creates the appropriate route based on the input parameters.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls will be overwritten
                         if `replace=True` option is provided, otherwise an error
                         will be raised.

        :keyword tuple(str) methods: http methods that this route could handle.
                                     if not provided, defaults to `GET`, `HEAD`
                                     and `OPTIONS`.

        :keyword callable view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule. pyrin
                               itself assumes the url rule as endpoint if not provided.

        :keyword bool login_required: specifies that this route could not be accessed
                                      if the requester has not a valid token.
                                      defaults to True if not provided.

        :keyword bool fresh_token: specifies that this route could not be accessed
                                   if the requester has not a valid fresh token.
                                   fresh token means a token that has been created by
                                   providing user credentials to server.
                                   defaults to False if not provided.

        :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                    to access this route's resource.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :raises RouteAuthenticationMismatchError: route authentication mismatch error.

        :rtype: RouteBase
        """

        route = self._create_route(rule, **options)
        if route is not None:
            return route

        login_required = options.get('login_required', True)
        fresh_token = options.get('fresh_token', False)

        if login_required is False and fresh_token is False:
            return PublicRoute(rule, **options)
        elif login_required is True and fresh_token is False:
            return ProtectedRoute(rule, **options)
        elif login_required is True and fresh_token is True:
            return FreshProtectedRoute(rule, **options)
        else:
            raise RouteAuthenticationMismatchError('"login_required={login}" and '
                                                   '"fresh_token={fresh}" in route '
                                                   '[{route}] are incompatible.'
                                                   .format(login=login_required,
                                                           fresh=fresh_token,
                                                           route=rule))

    def _create_route(self, rule, **options):
        """
        creates the appropriate route based on the input parameters.
        this method is intended to be overridden by subclasses to provided
        custom `RouteBase` types, it should always return a `RouteBase`
        object or `None`.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls will be overwritten
                         if `replace=True` option is provided, otherwise an error
                         will be raised.

        :keyword tuple(str) methods: http methods that this route could handle.
                                     if not provided, defaults to `GET`, `HEAD`
                                     and `OPTIONS`.

        :keyword callable view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule. pyrin
                               itself assumes the url rule as endpoint if not provided.

        :keyword bool login_required: specifies that this route could not be accessed
                                      if the requester has not a valid token.
                                      defaults to True if not provided.

        :keyword bool fresh_token: specifies that this route could not be accessed
                                   if the requester has not a valid fresh token.
                                   fresh token means a token that has been created by
                                   providing user credentials to server.
                                   defaults to False if not provided.

        :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                    to access this route's resource.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :rtype: RouteBase
        """

        return None

    def add_route(self, url, endpoint=None, view_func=None,
                  provide_automatic_options=None, **options):
        """
        connects a url rule. if a view_func is provided it will be registered with the endpoint.
        if there is another rule with the same url and `replace=True` option is provided,
        it will be replaced, otherwise an error will be raised.

        :param str url: the url rule as string.

        :param str endpoint: the endpoint for the registered url rule.
                             pyrin itself assumes the url rule as endpoint.

        :param callable view_func: the function to call when serving a request to the
                                   provided endpoint.

        :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                               added automatically.
                                               this can also be controlled by setting the
                                               `view_func.provide_automatic_options = False`
                                               before adding the rule.

        :keyword tuple(str) methods: http methods that this rule should handle.
                                     if not provided, defaults to `GET`.

        :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                    to access this route's resource.

        :keyword bool login_required: specifies that this route could not be accessed
                                      if the requester has not a valid token.
                                      defaults to True if not provided.

        :keyword bool fresh_token: specifies that this route could not be accessed
                                   if the requester has not a valid fresh token.
                                   fresh token means a token that has been created by
                                   providing user credentials to server.
                                   defaults to False if not provided.

        :keyword bool replace: specifies that this route must replace
                               any existing route with the same url or raise
                               an error if not provided. defaults to False.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :raises DuplicateRouteURLError: duplicate route url error.
        """

        application_services.add_url_rule(url, endpoint=endpoint, view_func=view_func,
                                          provide_automatic_options=provide_automatic_options,
                                          **options)



# -*- coding: utf-8 -*-
"""
packaging hooks module.
"""

from pyrin.core.context import Hook


class PackagingHookBase(Hook):
    """
    packaging hook base class.
    all packages that need to be hooked into packaging business must
    implement this class and register it in packaging hooks.
    """

    def __init__(self):
        """
        initializes an instance of PackagingHookBase.
        """

        super().__init__()

    def after_packages_loaded(self):
        """
        this method will be called after all application packages has been loaded.
        """
        pass



# -*- coding: utf-8 -*-
"""
protected route handler module.
"""

import pyrin.security.authorization.services as authorization_services
import pyrin.security.session.services as session_services

from pyrin.api.router.handlers.base import RouteBase
from pyrin.api.router.handlers.exceptions import FreshTokenRequiredError, PermissionTypeError
from pyrin.core.globals import _, LIST_TYPES
from pyrin.security.permission.base import PermissionBase


class ProtectedRoute(RouteBase):
    """
    protected route class.
    this class should be used to manage application protected api routes that need valid token.
    """

    def __init__(self, rule, **options):
        """
        initializes a new instance of ProtectedRoute.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls will be overwritten
                         if `replace=True` option is provided, otherwise an error
                         will be raised.

        :keyword tuple(str) methods: http methods that this route could handle.
                                     if not provided, defaults to `GET`, `HEAD`
                                     and `OPTIONS`.

        :keyword callable view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule. pyrin
                               itself assumes the rule as endpoint if not provided.

        :keyword dict defaults: an optional dict with defaults for other rules with the same
                                endpoint. this is a bit tricky but useful if you want
                                to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the rule
                                only matches for the `default_subdomain` of the map. if the map is
                                not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only
                                      for this rule. if not specified the `Map` setting is used.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you
                                  have resources on a subdomain or folder that are not
                                  handled by the WSGI application (like static data)

        :keyword Union[string, Callable] redirect_to: if given this must be either a string or
                                                      callable. in case of a callable it's
                                                      called with the url adapter that triggered
                                                      the match and the values of the url as
                                                      keyword arguments and has to return the
                                                      target for the redirect, otherwise it
                                                      has to be a string with placeholders
                                                      in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.
                               
        :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                    to access this route's resource.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises PermissionTypeError: permission type error.
        """

        super().__init__(rule, **options)

        self._permissions = options.get('permissions', ())
        if not isinstance(self._permissions, LIST_TYPES):
            self._permissions = (self._permissions,)

        if not all(isinstance(item, PermissionBase) for item in self._permissions):
            raise PermissionTypeError('All route permissions must be an '
                                      'instance of [{instance}].'
                                      .format(instance=PermissionBase))

    def _handle(self, inputs, **options):
        """
        handles the current route.
        routes which need to perform extra operations before
        view function execution, must override this method.

        :param dict inputs: view function inputs.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        self._authorize()

    def _authorize(self):
        """
        authorizes the route permissions for current user.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        user = session_services.get_current_user()
        authorization_services.authorize(user, self.get_permissions())

    def get_permissions(self):
        """
        gets all required permissions to access this route.

        :returns: tuple(PermissionBase)

        :rtype: tuple
        """

        return self._permissions


class FreshProtectedRoute(ProtectedRoute):
    """
    fresh protected route class.
    this class should be used to manage application protected api routes that
    need valid fresh token. fresh token means a token that has been created
    by providing user credentials to server.
    """

    def __init__(self, rule, **options):
        """
        initializes a new instance of FreshProtectedRoute.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls will be overwritten
                         if `replace=True` option is provided, otherwise an error
                         will be raised.

        :keyword tuple(str) methods: http methods that this route could handle.
                                     if not provided, defaults to `GET`, `HEAD`
                                     and `OPTIONS`.

        :keyword callable view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule. pyrin
                               itself assumes the rule as endpoint if not provided.

        :keyword dict defaults: an optional dict with defaults for other rules with the same
                                endpoint. this is a bit tricky but useful if you want
                                to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the rule
                                only matches for the `default_subdomain` of the map. if the map is
                                not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only
                                      for this rule. if not specified the `Map` setting is used.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you
                                  have resources on a subdomain or folder that are not
                                  handled by the WSGI application (like static data)

        :keyword Union[string, Callable] redirect_to: if given this must be either a string or
                                                      callable. in case of a callable it's
                                                      called with the url adapter that triggered
                                                      the match and the values of the url as
                                                      keyword arguments and has to return the
                                                      target for the redirect, otherwise it
                                                      has to be a string with placeholders
                                                      in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                    to access this route's resource.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises PermissionTypeError: permission type error.
        """

        super().__init__(rule, **options)

    def _authorize(self):
        """
        authorizes the route permissions for current user.
        also checks that the user has a fresh token.

        :raises FreshTokenRequiredError: fresh token required error.
        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        if not session_services.is_fresh():
            raise FreshTokenRequiredError(_('Fresh token is required to '
                                            'access the requested resource.'))

        super()._authorize()



# -*- coding: utf-8 -*-
"""
packaging manager module.
"""

import os
import inspect

from threading import Lock
from importlib import import_module

import pyrin.application.services as application_services
import pyrin.configuration.services as config_services
import pyrin.utils.configuration as config_utils
import pyrin.utils.path as path_utils

from pyrin.core.globals import LIST_TYPES
from pyrin.core.mixin import HookMixin
from pyrin.packaging import PackagingPackage
from pyrin.core.context import DTO, Manager
from pyrin.packaging.base import Package
from pyrin.packaging.hooks import PackagingHookBase
from pyrin.utils.custom_print import print_info, print_default
from pyrin.packaging.exceptions import InvalidPackageNameError, \
    ComponentModuleNotFoundError


class PackagingManager(Manager, HookMixin):
    """
    packaging manager class.
    """

    _hook_type = PackagingHookBase
    _lock = Lock()

    def __init__(self):
        """
        creates a new instance of PackagingManager.
        """

        super().__init__()

        # this flag indicates that application has been loaded.
        # it is required for environments in which server starts
        # multiple threads before application gets loaded.
        self._is_loaded = False

        self._pyrin_package_name = None

        # holds the loaded packages.
        self._loaded_packages = []

        # holds the name of disabled packages.
        self._disabled_packages = []

        # holds the instance of all loaded modules.
        # in the form of: {str module_name: Module module}
        self._loaded_modules = DTO()

        # configs will be filled from packaging config file.
        self._configs = DTO()

        # these will keep all loaded components for different
        # categories inside them. extended components in each
        # category are those that extending the exact component
        # of their parent.
        # in the form of: dict(str package_name: list[str] modules)
        self._pyrin_components = DTO()
        self._application_components = DTO()
        self._custom_components = DTO()
        self._test_components = DTO()
        self._extended_application_components = DTO()
        self._other_application_components = DTO()
        self._extended_test_components = DTO()
        self._other_test_components = DTO()

    def _create_config_file(self):
        """
        creates packaging config file in application settings path if not available.
        """

        config_services.create_config_file(PackagingPackage.CONFIG_STORE_NAMES[0],
                                           ignore_on_existed=True)

    def _load_configs(self):
        """
        loads packaging configs from application's settings directory.
        """

        self._configs.clear()
        configs = config_utils.load(self._get_config_file_path())
        self._configs = configs.get('general')

    def _get_config_file_path(self):
        """
        gets packaging config file path.

        it looks for file in top level application settings, but
        if not found it, it uses the file from default settings.
        we have to re-implement this here, because configuration
        services is not present yet to be used.

        :rtype: str
        """

        app_settings_directory = application_services.get_settings_path()
        pyrin_settings_directory = application_services.get_default_settings_path()
        config_file_name = '{store}.config'.format(store=PackagingPackage.CONFIG_STORE_NAMES[0])
        config_path = path_utils.get_first_available_file(app_settings_directory,
                                                          pyrin_settings_directory,
                                                          file_name=config_file_name)

        return config_path

    def _initialize_loaded_packages(self):
        """
        adds `pyrin.application` and `pyrin.packaging` into loaded packages.
        those packages will load immediately and will not be added
        to loaded packages through normal operations.
        """

        for package in self._configs.ignored_packages:
            if package in ('pyrin.application', 'pyrin.packaging'):
                if package not in self._loaded_packages:
                    self._loaded_packages.append(package)

    def _initialize(self):
        """
        initializes required data.
        """

        self._loaded_packages.clear()
        self._loaded_modules.clear()
        self._pyrin_components.clear()
        self._application_components.clear()
        self._custom_components.clear()
        self._test_components.clear()
        self._extended_application_components.clear()
        self._other_application_components.clear()
        self._extended_test_components.clear()
        self._other_test_components.clear()

        self._pyrin_package_name = path_utils.get_pyrin_main_package_name()
        self._load_configs()
        self._initialize_loaded_packages()

    def load_components(self, **options):
        """
        loads required packages and modules for application startup.
        """

        try:
            if self._is_loaded is True:
                return

            self._lock.acquire()

            if self._is_loaded is True:
                return

            self._initialize()

            print_info('Loading application components...')

            self._find_pyrin_loadable_components()
            self._find_other_loadable_components()

            self._load_components(self._pyrin_components, **options)
            self._load_components(self._extended_application_components, **options)
            self._load_components(self._other_application_components, **options)
            self._load_components(self._custom_components, **options)

            if self._configs.load_test_packages is True:
                self._load_components(self._extended_test_components, **options)
                self._load_components(self._other_test_components, **options)

            self._after_packages_loaded()

            print_info('Total of [{count}] packages loaded.'
                       .format(count=len(self._loaded_packages)))

            self._create_config_file()
            self._is_loaded = True
        finally:
            if self._lock.locked():
                self._lock.release()

    def _after_packages_loaded(self):
        """
        this method will call `after_packages_loaded` method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_packages_loaded()

    def load(self, module_name, **options):
        """
        loads the specified module.

        :param str module_name: full module name.
                                example module_name = `pyrin.application.decorators`.

        :rtype: Module
        """

        module = import_module(module_name)
        self._loaded_modules[module.__name__] = module

        return module

    def _load_component(self, package_name, module_names, component_name, **options):
        """
        loads the given component.

        :param str package_name: full package name to be loaded.
        :param list[str] module_names: full module names to be loaded.
        :param str component_name: component name of this package.

        :raises ComponentModuleNotFoundError: component module not found error.
        """

        self.load(package_name)

        # component module should be loaded first if available, in case of
        # any other module needed package services in top level objects.
        component_module = None
        if component_name is not None:
            component_module = self._merge_module_name(package_name, component_name)

        if component_module is not None and component_module in module_names:
            self.load(component_module, **options)
        elif component_module is not None and component_module not in module_names:
            raise ComponentModuleNotFoundError('Component module [{name}] not '
                                               'found in [{package}] package.'
                                               .format(name=component_module,
                                                       package=package_name))

        for module in module_names:
            if module != component_module:
                self.load(module, **options)

        self._loaded_packages.append(package_name)

        print_default('[{package}] package loaded. including [{module_count}] modules.'
                      .format(package=package_name,
                              module_count=len(module_names)))

    def _load_components(self, components, **options):
        """
        loads the given components considering their dependency on each other.

        :param dict components: full package names and their
                                modules to be loaded.

        :type components: dict(str package_name: list[str] modules)
        """

        # a dictionary containing all dependent package names and their respective modules.
        # in the form of {package_name: [modules]}.
        dependent_components = DTO()

        for package in components:
            package_class = self._get_package_class(package)

            # checking whether this package has any dependencies.
            # if so, check those dependencies have been loaded or not.
            # if not, then put this package into dependent_packages and
            # load it later. otherwise load it now.
            if (package_class is None or
                len(package_class.DEPENDS) == 0 or
                self._is_dependencies_loaded(package_class.DEPENDS) is True) and \
               self._is_parent_loaded(package) is True:

                instance = None
                if package_class is not None:
                    instance = package_class()
                    instance.load_configs(config_services)

                component_name = None
                if instance is not None:
                    component_name = instance.COMPONENT_NAME
                self._load_component(package, components[package], component_name, **options)
            else:
                dependent_components[package] = components[package]

        # now, go through dependent components if any, and try to load them.
        if len(dependent_components) > 0:
            self._load_components(dependent_components, **options)

    def _find_pyrin_loadable_components(self):
        """
        finds all package and module names that should
        be loaded from pyrin package.
        """

        pyrin_root_path = application_services.get_pyrin_root_path()
        pyrin_path = application_services.get_pyrin_main_package_path()
        self._find_loadable_components(pyrin_root_path, include=pyrin_path)

    def _find_other_loadable_components(self):
        """
        finds all package and module names that should
        be loaded from other packages like application and tests.
        """

        application_root_path = application_services.get_application_root_path()
        pyrin_path = application_services.get_pyrin_main_package_path()
        self._find_loadable_components(application_root_path, exclude=pyrin_path)

    def _find_loadable_components(self, root_path, include=None,
                                  exclude=None, **options):
        """
        finds all package and module names that should
        be loaded included in given root path.

        :param str root_path: root path to look for components inside it.

        :param Union[str, list[str]] include: specify full directory names inside the
                                              root path to just loop inside those.
                                              otherwise it loops in all available
                                              directories.

        :param Union[str, list[str]] exclude: specify full directory names inside the
                                              root path to ignore them. otherwise
                                              it loops in all available directories.
        """

        if include is not None and not isinstance(include, LIST_TYPES):
            include = [include]

        if exclude is not None and not isinstance(exclude, LIST_TYPES):
            exclude = [exclude]

        for root, directories, file_names in os.walk(root_path, followlinks=True):
            temp_dirs = list(directories)
            for single_dir in temp_dirs:
                visiting_path = os.path.abspath(os.path.join(root, single_dir))
                if self._should_visit(include, exclude, visiting_path) is False:
                    directories.remove(single_dir)

            for directory in directories:
                combined_path = os.path.join(root, directory)

                if not self._is_package(combined_path):
                    continue

                package_name = self._get_package_name(combined_path, root_path)
                if self._is_ignored_package(package_name):
                    continue

                package_class = self._get_package_class(package_name)
                if package_class is not None and package_class.ENABLED is False:
                    self._disabled_packages.append(package_name)
                    continue

                if self._is_disabled_package(package_name):
                    continue

                if self._is_pyrin_package(package_name):
                    self._pyrin_components[package_name] = []
                elif self._is_custom_package(package_name):
                    self._custom_components[package_name] = []
                elif self._is_test_package(package_name):
                    self._test_components[package_name] = []
                else:
                    self._application_components[package_name] = []

                files = os.listdir(combined_path)
                for file_name in files:
                    if not self._is_module(file_name):
                        continue

                    module_name = file_name.replace('.py', '')
                    full_module_name = self._get_module_name(package_name, module_name)
                    if self._is_ignored_module(full_module_name):
                        continue

                    if self._is_pyrin_module(full_module_name):
                        self._pyrin_components[package_name].append(full_module_name)
                    elif self._is_custom_module(full_module_name):
                        self._custom_components[package_name].append(full_module_name)
                    elif self._is_test_module(full_module_name):
                        self._test_components[package_name].append(full_module_name)
                    else:
                        self._application_components[package_name].append(full_module_name)

        self._detach_all()

    def _detach_all(self):
        """
        detaches all founded components into extended and other components.
        """

        self._extended_application_components, \
            self._other_application_components = self._detach_extended_packages(
                list(self._pyrin_components.keys()), self._application_components)

        test_base_components = self._application_components
        if len(self._application_components) <= 0:
            test_base_components = self._pyrin_components
        self._extended_test_components, \
            self._other_test_components = self._detach_extended_packages(
                list(test_base_components.keys()), self._test_components)

    def _detach_extended_packages(self, base_components, components):
        """
        detaches components which extend existing base components
        from those which are new components.

        :param list[str] base_components: base component names.

        :param dict components: components which some of
                                them extend base components.

        :type components: dict(str package_name: list[str] modules)

        :returns: tuple(extended_components, other_components)

        :type extended_components: dict(str package_name: list[str] modules)
        :type other_components: dict(str package_name: list[str] modules)

        :rtype: tuple
        """

        extended_components = DTO()
        other_components = DTO()
        if len(components) > 0:
            component_keys = list(components.keys())
            root_name = self._get_root_package(component_keys[0])
            base_names = [self._replace_root_package(item, root_name) for item in base_components]
            for package in components:
                if package in base_names:
                    extended_components[package] = components[package]
                else:
                    other_components[package] = components[package]

        return extended_components, other_components

    def _replace_root_package(self, old_package, root_name):
        """
        replaces root package name in old package with given root name.

        :param str old_package: old package fully qualified name.
                                in the form of: `pyrin.api`

        :param str root_name: root name to be put in old package name.
                              in the form of: `my_root`
        :rtype: str
        """

        parts = old_package.split('.')
        parts.pop(0)
        parts.insert(0, root_name)

        return '.'.join(parts)

    def _get_root_package(self, component_name):
        """
        gets the root package of given component.

        :param str component_name: full package or module name.

        :rtype: str
        """

        return path_utils.get_main_package_name(component_name)

    def _is_included(self, include, visiting_path):
        """
        returns a value indicating that the given
        visiting path is under include path.

        :param list[str] include: full directory names inside the
                                  root path to just loop inside those.
                                  otherwise it loops in all available
                                  directories.

        :param str visiting_path: full path which must be checked for inclusion.

        :rtype: bool
        """

        if include is None or len(include) <= 0:
            return True

        for path in include:
            if visiting_path.startswith(path):
                return True

        return False

    def _is_excluded(self, exclude, visiting_path):
        """
        returns a value indicating that the given
        visiting path is under exclude path.

        :param list[str] exclude: full directory names inside the
                                  root path to ignore them. otherwise
                                  it loops in all available directories.

        :param str visiting_path: full path which must be checked for exclusion.

        :rtype: bool
        """

        if exclude is None or len(exclude) <= 0:
            return False

        for path in exclude:
            if visiting_path.startswith(path):
                return True

        return False

    def _should_visit(self, include, exclude, visiting_path):
        """
        gets a value indicating that given path should be visited.

        :param list[str] include: full directory names inside the
                                  root path to just loop inside those.
                                  otherwise it loops in all available
                                  directories.

        :param list[str] exclude: full directory names inside the
                                  root path to ignore them. otherwise
                                  it loops in all available directories.

        :param str visiting_path: full path which must be checked for exclusion.

        :rtype: bool
        """

        return self._is_included(include, visiting_path) is True and \
            self._is_excluded(exclude, visiting_path) is False

    def _is_equal(self, condition, full_module_or_package):
        """
        gets a value indicating that given name is equal with given condition.

        :param str condition: condition to compare with name.
                              for example: `*.database`, then all values
                              that have database in their second part
                              and their parts count are equal to or greater
                              than condition parts count will be recognized as equal.

        :param str full_module_or_package: module or package name to be compared.
                                           for example: `pyrin.database`.

        :rtype: bool
        """

        condition_parts = condition.split('.')
        full_module_or_package_parts = full_module_or_package.split('.')

        if len(condition_parts) > len(full_module_or_package_parts):
            return False

        index = 0
        for item in condition_parts:
            if item != '*' and item != full_module_or_package_parts[index]:
                return False
            index = index + 1

        return True

    def _is_disabled_package(self, package_name):
        """
        gets a value indicating that given package should
        be considered as disabled based on its parents status.

        :param str package_name: full package name.
                                 example package_name = `pyrin.database`.

        :rtype: bool
        """

        for disabled in self._disabled_packages:
            if package_name.startswith(disabled) or self._is_equal(disabled, package_name):
                return True

        return False

    def _is_ignored_package(self, package_name):
        """
        gets a value indicating that given package should be ignored.

        :param str package_name: full package name.
                                 example package_name = `pyrin.database`.

        :rtype: bool
        """

        for ignored in self._configs.ignored_packages:
            if package_name.startswith(ignored) or self._is_equal(ignored, package_name):
                return True

        return False

    def _is_ignored_module(self, module_name):
        """
        gets a value indicating that given module should be ignored.

        :param str module_name: full module name.
                                example module_name = `pyrin.api.error_handlers`.

        :rtype: bool
        """

        for ignored in self._configs.ignored_modules:
            if module_name.endswith(ignored) or self._is_equal(ignored, module_name):
                return True

        return False

    def _is_pyrin_component(self, component_name):
        """
        gets a value indicating that given component is a pyrin component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        return self._contains(self._pyrin_package_name, component_name)

    def _is_pyrin_package(self, package_name):
        """
        gets a value indicating that given package is a pyrin package.

        :param str package_name: full package name.
                                 example package_name = 'pyrin.api'

        :rtype: bool
        """

        return self._is_pyrin_component(package_name)

    def _is_pyrin_module(self, module_name):
        """
        gets a value indicating that given module is a pyrin module.

        :param str module_name: full module name.
                                example module_name = 'pyrin.api.error_handlers'

        :rtype: bool
        """

        return self._is_pyrin_component(module_name)

    def _is_custom_component(self, component_name):
        """
        gets a value indicating that given component is a custom component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        for custom in self._configs.custom_packages:
            if self._contains(custom, component_name) is True:
                return True

        return False

    def _is_custom_package(self, package_name):
        """
        gets a value indicating that given package is a custom package.

        :param str package_name: full package name.
                                 example package_name = 'custom.api'

        :rtype: bool
        """

        return self._is_custom_component(package_name)

    def _is_custom_module(self, module_name):
        """
        gets a value indicating that given module is a custom module.

        :param str module_name: full module name.
                                example module_name = 'custom.api.error_handlers'

        :rtype: bool
        """

        return self._is_custom_component(module_name)

    def _is_test_component(self, component_name):
        """
        gets a value indicating that given component is a test component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        for test in self._configs.test_packages:
            if self._contains(test, component_name) is True:
                return True

        return False

    def _is_test_package(self, package_name):
        """
        gets a value indicating that given package is a test package.

        :param str package_name: full package name.
                                 example package_name = 'test.api'

        :rtype: bool
        """

        return self._is_test_component(package_name)

    def _is_test_module(self, module_name):
        """
        gets a value indicating that given module is a test module.

        :param str module_name: full module name.
                                example module_name = 'test.api.error_handlers'

        :rtype: bool
        """

        return self._is_test_component(module_name)

    def _contains(self, root, component_name):
        """
        gets a value indicating that given component
        qualified name, is belonging to the provided root.

        :param str root: root name that should be checked for component existence.
                         example root = `application.custom`

        :param str component_name: component name that should be
                                   checked for existence in root.
                                   example component_name = `application.custom.api`

        :rtype: bool
        """

        parts_component = component_name.split('.')
        parts_root = root.split('.')
        if len(parts_component) < len(parts_root):
            return False

        component_root = '.'.join(parts_component[0:len(parts_root)])

        return component_root == root

    def _get_package_name(self, path, root_path):
        """
        gets the full package name for provided path.

        :param str path: full path of package.
                         example path = `/home/src/pyrin/database`.

        :param str root_path: root path in which this path is located.
                              example root_path = `/home/src`

        :rtype: str
        """

        return path.replace(root_path, '').replace(os.path.sep, '.').lstrip('.')

    def _merge_module_name(self, package_name, component_name):
        """
        merges package and component name and gets the fully qualified module name.

        :param str package_name: package name.
                                 example package_name = `pyrin.database`.

        :param str component_name: component name.
                                   example component_name = `database.component`.

        :rtype: str
        """

        parts = component_name.split('.')
        return self._get_module_name(package_name, parts[-1])

    def _get_module_name(self, package_name, module_name):
        """
        gets the full module name.

        :param str package_name: package name.
                                 example package_name = `pyrin.database`.

        :param str module_name: module name.
                                example module_name = `api`.

        :rtype: str
        """

        return '{package}.{module}'.format(package=package_name, module=module_name)

    def _is_package(self, path):
        """
        gets a value indicating that given path belongs to a python package.
        it simply checks that `__init__` module exists or not.

        :param str path: full path of package.
                         example path = `/home/src/pyrin/database`.

        :rtype: bool
        """

        return self._has_module(path, '__init__')

    def _is_module(self, file_name):
        """
        gets a value indicating that given file is a standalone
        python module (excluding `__init__` module which belongs to package).
        it simply checks that file name ends with '.py' and not being `__init__.py`.

        :param str file_name: file name.
                              example file_name = `services.py`
        :rtype: bool
        """

        return file_name.endswith('.py') and '__init__.py' not in file_name

    def _has_module(self, path, module_name):
        """
        gets a value indicating that given module exists in specified path.

        :param str path: path to check module availability in it.
                         example path = `/home/src/pyrin/database`.

        :param str module_name: module name.
                                example module_name = `__init__`.

        :rtype: bool
        """

        return os.path.isfile(os.path.join(path, '{module}.py'.format(module=module_name)))

    def _get_package_class(self, package_name):
        """
        gets the package class implemented in given package if available, otherwise returns None.

        :param str package_name: full package name.
                                 example package_name = `pyrin.api`.

        :rtype: type
        """

        module = self.load(package_name)
        package_class = None

        for cls in module.__dict__.values():
            if inspect.isclass(cls) and cls is not Package and issubclass(cls, Package):
                package_class = cls
        return package_class

    def _is_dependencies_loaded(self, dependencies):
        """
        gets a value indicating that given dependencies has been already loaded.

        :param list[str] dependencies: full dependency names.
                                       example dependencies = `pyrin.logging`

        :rtype: bool
        """

        for dependency in dependencies:
            if dependency not in self._loaded_packages:
                return False

        return True

    def _is_parent_loaded(self, package_name):
        """
        gets a value indicating that given package's parent package has been loaded.

        :param str package_name: full package name.
                                 example package_name = `pyrin.encryption.handlers`

        :raises InvalidPackageNameError: invalid package name error.

        :rtype: bool
        """

        items = package_name.split('.')
        parent_package = None

        length = len(items)
        if length == 1:
            parent_package = items[0]
        elif length > 1:
            parent_package = '.'.join(items[:-1])
        else:
            raise InvalidPackageNameError('Input package name [{package_name}] is invalid.'
                                          .format(package_name=package_name))

        # application root packages like `pyrin`, has no
        # parent so it should always return `True` for them.
        if parent_package == package_name:
            return True

        return self._is_dependencies_loaded([parent_package])



# -*- coding: utf-8 -*-
"""
migrations env module.
"""

from os import path

from sqlalchemy import engine_from_config
from sqlalchemy import pool
from alembic import context

import pyrin.database.migration.services as migration_services
import pyrin.application.services as application_services
import pyrin.globalization.datetime.services as datetime_services
import pyrin.configuration.services as config_services
import pyrin.logging.services as logging_services

from pyrin.utils.custom_print import print_default
from pyrin.utils.environment import set_python_path

set_python_path('.')

from APPLICATION_MODULE import APPLICATION_CLASS


app_instance = APPLICATION_CLASS(scripting_mode=True)

USE_TWOPHASE = config_services.get('alembic', 'alembic', 'use_twophase')

LOGGER = logging_services.get_logger('alembic')

# gather section names referring to different
# databases. these are named according to bind names
# in the alembic.config file. default database must
# always be referenced by 'default' key.
db_names = config_services.get('alembic', 'alembic', 'databases')

# getting timezone from alembic.config to set in file names.
timezone = config_services.get('alembic', 'alembic', 'timezone')

# keeps a collection of connection bind names and urls.
# default engine bind name is always default.
connection_urls = migration_services.get_connection_urls()

# add your model's 'MetaData' objects here
# for 'autogenerate' support. these must be set
# up to hold just those tables targeting a
# particular database. table.tometadata() may be
# helpful here in case a 'copy' of
# a metadata is needed.
# from myapp import mymodel
# target_metadata = {
#       'default':mymodel.metadata1,
#       'engine2':mymodel.metadata2
# }
target_metadata = migration_services.get_bind_name_to_metadata_map()


def print_message(message):
    """
    prints the given message into stdout.

    :param str message: message to be printed.
    """

    print_default(message, force=True)


def run_migrations_offline():
    """
    run migrations in 'offline' mode.

    this configures the context with just a url
    and not an engine, though an engine is acceptable
    here as well. by skipping the engine creation
    we don't even need a dbapi to be available.

    calls to context.execute() here emit the given string to the
    script output.
    """

    # for the --sql use case, run migrations for each URL into
    # individual files.

    engines = {}
    for name in db_names:
        engines[name] = rec = {}
        rec['url'] = connection_urls.get(name)
        rec['engine'] = engine_from_config(
            dict(url=connection_urls.get(name)),
            prefix="",
            poolclass=pool.NullPool,
        )

    timestamp = datetime_services.get_current_timestamp(date_sep=None,
                                                        main_sep=None,
                                                        time_sep=None,
                                                        timezone=timezone)
    for name, rec in engines.items():
        engine = rec['engine']
        rec['connection'] = conn = engine.connect()
        print_message('Migrating database [{name}]'.format(name=name))

        migrations_path = application_services.get_migrations_path()
        file_ = 'sql/{timestamp}_{name}.sql'.format(timestamp=timestamp,
                                                    name=name)
        full_migrations_file_path = path.join(migrations_path, file_)

        message = 'Writing output to [{file_name}]'.format(file_name=full_migrations_file_path)
        LOGGER.debug(message)
        print_message(message)

        with open(full_migrations_file_path, 'w') as buffer:
            context.configure(
                url=rec['url'],
                connection=rec['connection'],
                output_buffer=buffer,
                target_metadata=target_metadata.get(name),
                literal_binds=True,
                as_sql=True,
                dialect_opts={'paramstyle': 'named'},
                compare_type=config_services.get('alembic', 'alembic', 'compare_type'),
                compare_server_default=config_services.get('alembic', 'alembic',
                                                           'compare_server_default'),
                render_as_batch=config_services.get('alembic', 'alembic', 'render_as_batch'),
                include_schemas=config_services.get('alembic', 'alembic', 'include_schemas'),
            )
            with context.begin_transaction():
                context.run_migrations(engine_name=name)


def run_migrations_online():
    """
    run migrations in 'online' mode.

    in this scenario we need to create an engine
    and associate a connection with the context.
    """

    # for the direct-to-db use case, start a transaction on all
    # engines, then run all migrations, then commit all transactions.

    engines = {}
    for name in db_names:
        engines[name] = rec = {}
        rec['engine'] = engine_from_config(
            dict(url=connection_urls.get(name)),
            prefix="",
            poolclass=pool.NullPool,
        )

    for name, rec in engines.items():
        engine = rec['engine']
        rec['connection'] = conn = engine.connect()

        if USE_TWOPHASE:
            rec['transaction'] = conn.begin_twophase()
        else:
            rec['transaction'] = conn.begin()

    try:
        for name, rec in engines.items():
            message = 'Migrating database [{name}]'.format(name=name)
            LOGGER.debug(message)
            print_message(message)
            context.configure(
                connection=rec['connection'],
                upgrade_token='%s_upgrades' % name,
                downgrade_token='%s_downgrades' % name,
                target_metadata=target_metadata.get(name),
                user_module_prefix='pyrin.database.migration.types.',
                compare_type=config_services.get('alembic', 'alembic', 'compare_type'),
                compare_server_default=config_services.get('alembic', 'alembic',
                                                           'compare_server_default'),
                render_as_batch=config_services.get('alembic', 'alembic', 'render_as_batch'),
                include_schemas=config_services.get('alembic', 'alembic', 'include_schemas'),
            )
            context.run_migrations(engine_name=name)

        if USE_TWOPHASE:
            for rec in engines.values():
                rec['transaction'].prepare()

        for rec in engines.values():
            rec['transaction'].commit()

    except Exception:
        for rec in engines.values():
            rec['transaction'].rollback()
        raise
    finally:
        for rec in engines.values():
            rec['connection'].close()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()
